# Copyright 2021-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""This file is used to define the DataLoader."""
import csv
import enum
import os
import json
import re
from collections import namedtuple
from pathlib import Path

from google.protobuf.message import DecodeError

from mindinsight.debugger.common.exceptions.exceptions import DebuggerParamValueError, RankDirNotFound, \
    DebuggerJsonFileParseError, DebuggerNodeTooLarge
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.common.utils import DumpSettings, is_valid_rank_dir_name
from mindinsight.domain.graph.proto.ms_graph_pb2 import ModelProto

RankDir = namedtuple("rank_dir", ["rank_id", "path"])


@enum.unique
class DumpTarget(enum.Enum):
    """Define the status of reply."""
    FULL = 0
    INPUT_ONLY = 1
    OUTPUT_ONLY = 2


@enum.unique
class FileFormat(enum.Enum):
    """Define the file format."""
    BIN = "bin"
    NPY = "npy"


class DataLoader:
    """The DataLoader object provides interface to load graphs and device information from base_dir."""

    DUMP_METADATA = '.dump_metadata'
    EXECUTION_ORDER = 'execution_order'
    GRAPHS = 'graphs'

    def __init__(self, base_dir):
        self._debugger_base_dir = Path(base_dir).absolute()
        # list of RankDir objects
        self._rank_dirs = []
        # flag for whether the data is from sync dump or async dump.
        self._is_sync = False
        self._device_target = "Ascend"
        self._file_format = ""
        self._net_name = ""
        self._dump_type = 0
        self._initialize()

    @property
    def file_format(self):
        """The property of file format."""
        return self._file_format

    @property
    def device_target(self):
        """The property of device target."""
        return self._device_target

    @property
    def dump_dir(self):
        """The property of dump directory."""
        return str(self._debugger_base_dir)

    @property
    def dump_target(self):
        """The property of dump target."""
        return DumpTarget(self._dump_type)

    @property
    def rank_dirs(self):
        """The property of RankDir."""
        return self._rank_dirs

    def _initialize(self):
        """Initialize the data_mode and net_dir of DataLoader."""
        self._load_rank_dirs()
        if not self._rank_dirs:
            log.error("No rank directory found under %s", str(self._debugger_base_dir))
            raise RankDirNotFound(str(self._debugger_base_dir))
        rank_dir = self._rank_dirs[0].path
        dump_config = self._load_json_file(rank_dir / self.DUMP_METADATA / 'data_dump.json')

        def _parse_common_settings():
            nonlocal dump_config
            common_settings = dump_config.get(DumpSettings.COMMON_DUMP_SETTINGS.value, {})
            try:
                self._net_name = common_settings['net_name']
                self._dump_type = common_settings['input_output']
                self._file_format = common_settings.get('file_format', FileFormat.BIN.value)
            except KeyError:
                raise DebuggerJsonFileParseError("data_dump.json")

        def _set_dump_mode_and_device_target():
            nonlocal dump_config
            config_json = self._load_json_file(rank_dir / self.DUMP_METADATA / 'config.json')
            self._device_target = config_json.get('device_target', 'Ascend')
            if self._device_target == 'GPU' or dump_config.get(DumpSettings.E2E_DUMP_SETTINGS.value) and \
                    dump_config[DumpSettings.E2E_DUMP_SETTINGS.value]['enable']:
                self._is_sync = True
            else:
                self._is_sync = False

        _parse_common_settings()
        _set_dump_mode_and_device_target()

    def _load_rank_dirs(self):
        """Load rank directories."""
        self._rank_dirs = []
        rank_dirs = self._debugger_base_dir.glob('rank_*')
        for rank_dir in rank_dirs:
            if not rank_dir.is_dir() or not is_valid_rank_dir_name(rank_dir.name) \
                    or not os.path.exists(os.path.join(rank_dir, ".dump_metadata")):
                continue
            rank_id = int(rank_dir.name.split('_', 1)[-1])
            self._rank_dirs.append(RankDir(rank_id, rank_dir))
        if self._rank_dirs:
            self._rank_dirs.sort(key=lambda x: x.rank_id)

    def load_graphs(self, threshold=None):
        """
        Load graphs from the debugger_base_dir.

        Args:
            threshold (int): Node limit during load graphs.

        Returns:
            list, list of graph protos from all ranks. Each item is like:
                {'rank_id': int,
                'graph_protos': [GraphProto]}
        """
        graph_loader = GraphLoader(threshold)
        res = []
        for rank_dir in self._rank_dirs:
            rank_id, rank_path = rank_dir.rank_id, rank_dir.path
            graphs_dir = rank_path / self.GRAPHS
            if not graphs_dir.is_dir():
                log.debug("Directory '%s' doesn't exist.", graphs_dir)
                res.append({'rank_id': rank_id, 'graph_protos': []})
                continue
            graph_protos = graph_loader.get_graph_protos_from_dir(graphs_dir)
            res.append({'rank_id': rank_id, 'graph_protos': graph_protos})

        return res

    def load_device_info(self):
        """Load device_info from dump path."""
        device_info = {}
        if not self._rank_dirs:
            log.info("No rank directory found under dump path.")
            return device_info
        rank_dir = self._rank_dirs[0].path
        hccl_json = self._load_json_file(rank_dir / '.dump_metadata' / 'hccl.json')
        if hccl_json.get('server_list'):
            device_info = {'device_target': self._device_target, 'server_list': hccl_json['server_list']}
        else:
            log.info("Server List info is missing. Set device id same with rank id as default.")
            devices = []
            for rank_dir in self._rank_dirs:
                rank_id = rank_dir.rank_id
                devices.append({'device_id': str(rank_id), 'rank_id': str(rank_id)})
            device_info = {'device_target': self._device_target,
                           'server_list': [{'server_id': 'localhost', 'device': devices}]}
        return device_info

    @staticmethod
    def _load_json_file(file):
        """
        Load json file content.

        Args:
            file (Path): The Path object.

        Returns:
            dict, the json content.
        """
        if not file.is_file():
            log.info("File <%s> is missing.", str(file))
            return {}
        with file.open() as handler:
            try:
                return json.load(handler)
            except json.decoder.JSONDecodeError as err:
                log.warning("Failed to load json file %s. %s", str(file), str(err))
                return {}

    def load_dumped_step(self, rank_id=None):
        """
        Load dumped step number in the directory.

        Args:
            rank_id (int): The specified rank id to query dumped steps. If None, load all step number.

        Returns:
            dict, the dumped step info among all devices. The key is rank_id,
            the value is like Dict[int, set[int]].
        """
        def _load_iters_per_rank(net_dir):
            res = {}
            for graph_dir in net_dir.iterdir():
                graph_id_str = graph_dir.name
                if not graph_id_str.isdigit():
                    log.warning("Invalid graph dir under net dir:%s", str(net_dir))
                    continue
                res[int(graph_id_str)] = _load_iters_per_graph(graph_dir)
            return res

        def _load_iters_per_graph(graph_dir):
            iters = set()
            for iteration_dir in graph_dir.iterdir():
                iter_id_str = iteration_dir.name
                if not iter_id_str.isdigit():
                    log.debug("Ignore iteration dir: %s", str(iteration_dir))
                    continue
                iters.add(int(iter_id_str))
            return iters

        log.debug("Start to load dumped steps.")
        step_num = {}
        rank_dirs = [self.get_rank_dir(rank_id)] if rank_id is not None else self._rank_dirs
        for rank_dir in rank_dirs:
            rank_id, rank_path = rank_dir.rank_id, rank_dir.path
            net_path = rank_path / self._net_name
            if not net_path.is_dir():
                log.info("No net directory under rank dir: %s", str(rank_dir))
                continue
            step_num[rank_id] = _load_iters_per_rank(net_path)

        return step_num

    def get_config_json_data(self):
        rank_dir = self._rank_dirs[0].path
        return self._load_json_file(rank_dir / self.DUMP_METADATA / 'config.json')

    def get_rank_dir(self, rank_id):
        """
        Get the rank directory according to rank_id.

        Args:
            rank_id (int): The rank id.

        Returns:
            RankDir, the rank dir info.
        """
        rank_path = self._debugger_base_dir / f'rank_{rank_id}'
        if rank_path.is_dir():
            return RankDir(rank_id, rank_path)
        log.error("No rank directory found.")
        raise DebuggerParamValueError("Invalid rank_id.")

    def get_step_iter(self, rank_id=None, step=None, root_graph_id=None):
        """Get the generator of step path."""
        step_pattern = '[0-9]*' if step is None else step
        graph_pattern = '[0-9]*' if root_graph_id is None else root_graph_id
        if rank_id is None:
            rank_dirs = self._rank_dirs
        else:
            rank_dirs = [self.get_rank_dir(rank_id)]
        for rank_dir in rank_dirs:
            step_dirs = rank_dir.path.glob(f'{self._net_name}/{graph_pattern}/{step_pattern}/')
            for step_dir in step_dirs:
                yield step_dir

    def get_dump_dir(self):
        """Get graph_name directory of the data."""
        return str(self._debugger_base_dir)

    def get_sync_flag(self):
        """Get the sync flag of the data."""
        return self._is_sync

    def get_net_name(self):
        """Get net_name of the data."""
        return self._net_name

    def load_graph_history(self):
        """
        Load graph history of all devices.

        Returns:
            dict, graph history. The key is rank id, the value is Dict[int, set[int]].
        """
        log.debug("Start to load graph history.")
        all_graph_history = {}
        for rank_dir in self._rank_dirs:
            execution_order_dir = rank_dir.path / self.EXECUTION_ORDER
            if not execution_order_dir.is_dir():
                log.warning("Execution order directory %s doesn't exist.", str(execution_order_dir))
                all_graph_history[rank_dir.rank_id] = {}
                continue
            graph_history = {}
            history_pattern = re.compile(r"ms_global_execution_order_graph_(?P<graph_id>\d+).csv")
            for item in execution_order_dir.iterdir():
                re_match = re.match(history_pattern, item.name)
                if re_match:
                    history = read_graph_history(item)
                    graph_history[int(re_match.group('graph_id'))] = history
            all_graph_history[rank_dir.rank_id] = graph_history
        return all_graph_history

    def has_data(self, step, rank_id=None):
        """
        Check if the step has data dumped.

        Args:
            step (int): Step id to check.
            rank_id (int): The rank id of the step. If None, check same step in all ranks.

        Returns:
            bool, has data or not.
        """
        rank_dirs = [self.get_rank_dir(rank_id)] if rank_id is not None else self._rank_dirs
        for rank_dir in rank_dirs:
            rank_id, rank_path = rank_dir.rank_id, rank_dir.path
            net_path = rank_path / self._net_name
            if not net_path.is_dir():
                log.info("No net directory under rank dir: %s", str(rank_dir))
                return False
            for graph_dir in net_path.iterdir():
                target_step_dir = graph_dir / str(step)
                if target_step_dir.is_dir():
                    return True
        return False


def read_graph_history(file_path):
    """
    Load graph history from file.

    Args:
        file_path (str): File path contains the graph execution order info.

    Returns:
        set[int], the executed iteration ids of each graph.
    """
    with open(file_path, 'r') as handler:
        csv_reader = csv.reader(handler)
        history = set(int(row[0]) for row in csv_reader)
        return history


class GraphLoader:
    """
    Load graph proto object in control.

    Args:
        threshold (int): The node limit for load graphs. If None, there is no limit.
            Default: None.

    Raises:
        DebuggerNodeTooLarge: When load new graph, the total node count is over node threshold.
    """
    def __init__(self, threshold=None):
        self._node_limit = threshold
        # node number loaded currently
        self._total_node_num = 0

    @staticmethod
    def load_graph_from_file(graph_file_path):
        """
        Load graph from file.

        Args:
            graph_file_path (Path): Graph file path.

        Returns:
            GraphProto, the parsed GraphProto object.
        """
        with graph_file_path.open('rb') as file_handler:
            model_bytes = file_handler.read()
            model = ModelProto.FromString(model_bytes)
            graph = model.graph

        return graph

    def _add_node_count(self, graph_proto):
        """Accumulate node count."""
        if self._node_limit is None:
            return
        node_cnt = len(graph_proto.parameters) + len(graph_proto.const_vals) + len(graph_proto.node)
        self._total_node_num += node_cnt
        if self._total_node_num > self._node_limit:
            raise DebuggerNodeTooLarge(self._node_limit, self._total_node_num)

    def get_graph_protos_from_dir(self, graphs_dir):
        """
        Get graph from graph directory.

        Args:
            graphs_dir (Union[Path, str]): The path of graph directory.

        Returns:
            list, list of 'GraphProto' object.
        """
        graphs_dir = graphs_dir if isinstance(graphs_dir, Path) else Path(graphs_dir)
        graph_protos = []
        pre_file_name = "ms_output_trace_code_graph_"
        for file_in_device in graphs_dir.iterdir():
            file_name = file_in_device.name
            if file_name.startswith(pre_file_name) and file_name.endswith(".pb"):
                try:
                    graph_proto = self.load_graph_from_file(file_in_device)
                except DecodeError:
                    log.warning("Load graph failed. The graph file is invalid.")
                    return []
                self._add_node_count(graph_proto)
                graph_protos.append(graph_proto)
        return graph_protos



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Exception module."""

from importlib import import_module
from mindinsight.utils.constant import MindInsightModules, GeneralErrors


class MindInsightException(Exception):
    """
    Base class for MindInsight exception.

    Examples:
        >>> raise MindInsightException(GeneralErrors.PATH_NOT_EXISTS_ERROR, 'path not exists')
        >>> raise MindInsightException(DataVisualErrors.CUSTOMIZED_ERROR, 'datavisual error message')
    """

    RUNTIME = 1
    TYPE = 1
    LEVEL = 0
    SYSID = 42

    def __init__(self, error, message, http_code=400):
        """
        Initialization of MindInsightException.

        Args:
            error (Enum): Error value for specified case.
            message (str): Description for exception.
            http_code (int): Http code for exception. Default is 400.
        """
        if isinstance(message, str):
            message = ' '.join(message.split())
        super(MindInsightException, self).__init__(message)
        self.error = error
        self.message = message
        self.http_code = http_code

    def parse_module(self):
        """
        Parse module according to error enum class.

        Note:
            Each enum value, excluding GENERAL, has an Errors class name starting with the enum value
            in Camel-Case referring to specific module.

        Returns:
            Enum, module for specified error.
        """

        module = None
        constant = import_module('mindinsight.utils.constant')
        errors_names = [item for item in dir(constant) if item.endswith('Errors')]
        for name in errors_names:
            errors_cls = getattr(constant, name)
            if isinstance(self.error, errors_cls):
                key = name[:-len('Errors')].upper()
                module = getattr(MindInsightModules, key, None)
                break

        return module

    @property
    def error_code(self):
        """
        Transform exception no to MindInsight error code.

        code compose(4bytes):
        runtime 2bits, type 2bits, level 3bits, sysid 8bits, modid 5bits, value 12bits.

        num = ((0xFF & runtime) << 30) \
                | ((0xFF & type) << 28) \
                | ((0xFF & level) << 25) \
                | ((0xFF & sysid) << 17) \
                | ((0xFF & modid) << 12) \
                | (0x0FFF & value)

        Returns:
            str, Hex string representing the composed MindInsight error code.
        """

        module = self.parse_module()
        if not module:
            raise UnknownError('Unknown module for {}.'.format(self.error))

        num = (((0xFF & self.RUNTIME) << 30)
               | ((0xFF & self.TYPE) << 28)
               | ((0xFF & self.LEVEL) << 25)
               | ((0xFF & self.SYSID) << 17)
               | ((0xFF & module.value) << 12)
               | (0x0FFF & self.error.value))

        return hex(num)[2:].zfill(8).upper()

    def __str__(self):
        return '[{}] code: {}, msg: {}'.format(self.__class__.__name__, self.error_code, self.message)


class ParamValueError(MindInsightException):
    """Request param value error."""
    def __init__(self, error_detail=None):
        if error_detail is None:
            error_msg = "Invalid parameter value."
        else:
            error_msg = 'Invalid parameter value. {}'.format(error_detail)
        super(ParamValueError, self).__init__(
            GeneralErrors.PARAM_VALUE_ERROR,
            error_msg,
            http_code=400)


class ParamTypeError(MindInsightException):
    """Request param type error."""
    def __init__(self, param_name, expected_type):
        error_msg = "Invalid parameter type. '{}' expect {} type.".format(param_name, expected_type)
        super(ParamTypeError, self).__init__(
            GeneralErrors.PARAM_TYPE_ERROR,
            error_msg,
            http_code=400)


class ParamMissError(MindInsightException):
    """Missing param error."""
    def __init__(self, param_name):
        error_msg = "Param missing. '{}' is required.".format(param_name)
        super(ParamMissError, self).__init__(
            GeneralErrors.PARAM_MISSING_ERROR,
            error_msg,
            http_code=400)


class PathNotExistError(MindInsightException):
    """Raised when specified path do not exist."""
    def __init__(self, error_detail):
        """Initialize PathNotExistError."""
        error_msg = 'Specified path does not exist. Detail: {}'.format(error_detail)
        super(PathNotExistError, self).__init__(
            GeneralErrors.PATH_NOT_EXISTS_ERROR,
            error_msg,
            http_code=400)


class FileSystemPermissionError(MindInsightException):
    """Can not access file or dir."""
    def __init__(self, error_detail):
        error_msg = 'File or dir access failed. Detail: {}'.format(error_detail)
        super(FileSystemPermissionError, self).__init__(
            GeneralErrors.FILE_SYSTEM_PERMISSION_ERROR,
            error_msg,
            http_code=400)


class PortReuseException(MindInsightException):
    """Port not available error.."""
    def __init__(self, error_detail):
        error_msg = '{}'.format(error_detail)
        super(PortReuseException, self).__init__(
            GeneralErrors.PORT_NOT_AVAILABLE_ERROR,
            error_msg,
            http_code=400)


class PortNotAvailableError(MindInsightException):
    """Port not available error.."""
    def __init__(self, error_detail):
        error_msg = 'Port not available error. Detail: {}'.format(error_detail)
        super(PortNotAvailableError, self).__init__(
            GeneralErrors.PORT_NOT_AVAILABLE_ERROR,
            error_msg,
            http_code=400)


class UnknownError(MindInsightException):
    """Unknown error."""
    def __init__(self, error_msg):
        super(UnknownError, self).__init__(
            GeneralErrors.UNKNOWN_ERROR,
            error_msg,
            http_code=500)


class UrlDecodeError(MindInsightException):
    """Url decoding failed"""
    def __init__(self, error_detail):
        error_msg = f"Url decode failed. Detail: {error_detail}"
        super(UrlDecodeError, self).__init__(GeneralErrors.URL_DECODE_ERROR,
                                             error_msg,
                                             http_code=400)


class MindInsightMemoryError(MindInsightException):
    """Unknown error."""
    def __init__(self, error_msg):
        super(MindInsightMemoryError, self).__init__(
            GeneralErrors.MEMORY_ERROR,
            error_msg,
            http_code=400)

class SettingValueError(MindInsightException):
    """Setting value error."""
    def __init__(self, error_msg):
        super(SettingValueError, self).__init__(
            GeneralErrors.SETTING_VALUE_ERROR,
            error_msg,
            http_code=400)


class VersionNotMatchError(MindInsightException):
    """Version not match error."""
    def __init__(self, error_msg):
        super(VersionNotMatchError, self).__init__(
            GeneralErrors.PARAM_TYPE_ERROR,
            error_msg,
            http_code=400)



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Datavisual."""

from mindinsight.backend.datavisual.static_resource_api import init_module as static_init_module
from mindinsight.backend.datavisual.task_manager_api import init_module as task_init_module
from mindinsight.backend.datavisual.train_visual_api import init_module as train_init_module


def init_module(app):
    """
    Interface to init module.

    Args:
        app (Flask): An instance of Flask.

    """
    static_init_module(app)
    task_init_module(app)
    train_init_module(app)



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The StepTraceAnalyser analyser class."""
import csv
import json
import os

from mindinsight.datavisual.utils.tools import to_int
from mindinsight.profiler.analyser.base_analyser import BaseAnalyser
from mindinsight.profiler.common.exceptions.exceptions import ProfilerParamValueErrorException, \
    ProfilerFileNotFoundException, StepNumNotSupportedException, ProfilerRawFileException
from mindinsight.profiler.common.log import logger as log
from mindinsight.profiler.common.util import query_latest_trace_time_file, get_field_value, \
    get_summary_for_step_trace, to_millisecond
from mindinsight.profiler.common.validator.validate_path import validate_and_normalize_path


class StepTraceAnalyser(BaseAnalyser):
    """The analyser for analyzing training steps."""

    _col_names = []
    _attr_ui_name = 'name'
    _attr_ui_start = 'start'
    _attr_ui_duration = 'duration'
    _point_info = {}

    @property
    def summary(self):
        """The property of summary info."""

        summary = get_summary_for_step_trace(self._data[-1], self.__column__)
        summary['total_steps'] = self._size
        return summary

    def point_info(self, step_id):
        """The property of point info."""
        if isinstance(self._point_info, list):
            return self._point_info[step_id-1]
        return self._point_info

    def query(self, condition=None):
        """
        Query data according to the condition.

        Args:
            condition (dict): The search condition, only contains `filter_condition` parameter.
                Default: None.

        Returns:
            dict, the result after filtered, sorted and grouped.
        """
        if condition is None:
            condition = {}
        filter_condition = condition.get('filter_condition', {})
        log.info("Receive query request. %s", filter_condition)
        self._validate_filter_condition(filter_condition)
        self._result = {'size': self._size}
        self._filter(filter_condition)

        return self._result

    def query_for_all_reduce(self):
        """
        Query for all reduce info.

        Returns:
            list[dict], reduce information. Each item is the reduce info for one step.
            The reduce info is format like:
            {stream_id: List[Tuple(start_point, end_point, duration, field_name)]}.
        """
        reduce_infos = []
        for row_info in self._data[:-1]:
            row_info_dict = self._get_info_dict_from_row_data(row_info, 'systime')
            reduce_info = self._sort_reduce_by_time(row_info_dict)
            if reduce_info:
                reduce_infos.extend(reduce_info)

        return reduce_infos

    def _load(self):
        """Load data according to the parsed AICORE operator types file."""
        file_path = query_latest_trace_time_file(self._profiling_dir, self._device_id)
        if not file_path:
            log.error("Failed to find parsed trace time file.")
            raise ProfilerFileNotFoundException('parsed step trace time file')
        file_path = validate_and_normalize_path(
            file_path, raise_key="Invalid latest_trace_trace_time file path.")
        with open(file_path, 'r') as handle:
            csv_reader = csv.reader(handle)
            self.__column__ = next(csv_reader)
            self._data = list(csv_reader)
        self._size = len(self._data) - 1
        self._display_col_names = self._col_names[:]
        self._load_point_info()

    def _load_point_info(self):
        """Load point info."""
        file_path = os.path.join(self._profiling_dir, 'step_trace_point_info.json')
        file_path = validate_and_normalize_path(
            file_path, raise_key="Invalid step_trace_point_info file path.")

        # If step_trace_point_info_{self._device_id}.json file exist, load this file.
        file_path_new = os.path.join(self._profiling_dir, f'step_trace_point_info_{self._device_id}.json')
        file_path_new = validate_and_normalize_path(
            file_path_new, raise_key="Invalid step_trace_point_info file path.")
        if os.path.isfile(file_path_new):
            file_path = file_path_new

        if os.path.isfile(file_path):
            with open(file_path, 'r', encoding='utf-8') as file:
                try:
                    self._point_info = json.load(file)
                except (json.JSONDecodeError, TypeError) as err:
                    log.exception(err)
                    raise ProfilerRawFileException('Fail to parse point info file.')

    def _filter(self, filter_condition):
        """
        Filter the profiling data according to the filter condition.

        Args:
            filter_condition (dict): The filter condition.

                - mode (str): The kind of information. `step` return the info about specific
                    step. `proc` return the info about specific field in parsed trace file.

                - step_id (int): The selected step_id. If not given, it means all steps is required.
                    If the value is 0, it means average info for all steps except the first is
                    required.

                - proc_name (str): The selected field name.

                - time_type (str): The value type. `systime` keeps the original value.
                    `realtime` transforms the value in millisecond. Default: `realtime`.
        """
        mode = filter_condition.get('mode', 'step')
        if mode == 'step':
            self._get_step_details(step_id=filter_condition.get('step_id'),
                                   time_type=filter_condition.get('time_type', 'realtime'))
        else:
            self._get_proc_details(step_id=filter_condition.get('step_id'),
                                   proc_name=filter_condition.get('proc_name'),
                                   time_type=filter_condition.get('time_type', 'realtime'))

    def _construct_time_point(self, name, start, duration):
        """Construct time point."""
        point = {}
        if start >= 0 and duration >= 0:
            point = {
                self._attr_ui_name: name,
                self._attr_ui_start: round(start, 4),
                self._attr_ui_duration: round(duration, 4)
            }
        else:
            log.warning("Not invalid point info: "
                        "name: %s, start: %s, duration: %s", name, start, duration)
        return point

    def _get_step_details(self, step_id, time_type='realtime'):
        """
        Get step trace info for selected step and save the result.

        Args:
            step_id (int): The selected step_id. If the value is 0, it means average info
                for all steps except the first is required.
            time_type (str): The value type. `systime` keeps the original value.
                `realtime` transforms the value in millisecond. Default: `realtime`.
        """
        if step_id is None:
            step_id = 0
        row_info = self._data[step_id - 1]
        row_info_dict = self._get_info_dict_from_row_data(row_info, time_type)
        # first line only contains total time
        first_line = [self._construct_time_point('', 0, row_info_dict.get('total', 0))]
        # second line contains iteration_interval, fp_and_bp and tail
        second_line = self._get_main_proc_points(row_info_dict)
        # construct reduces lines
        reduce_lines = self._construct_reduce_lines(row_info_dict)

        graph = [first_line, second_line]
        graph.extend(reduce_lines)
        self._result['training_trace_graph'] = graph

    def _get_info_dict_from_row_data(self, row_info, time_type):
        """
        Get step info in dict format.

        Args:
            row_info (list[str]): Step info, the value is corresponding to `__column__`.
            time_type (str): The value type. `systime` keeps the original value.
                `realtime` transforms the value in millisecond. Default: `realtime`.

        Returns:
            dict, step trace information. The key is in `__column__`.
        """
        row_info_dict = {}
        for key, value in zip(self.__column__, row_info):
            if key == 'step_num':
                continue
            value = to_int(value, key)
            row_info_dict[key] = to_millisecond(value) if time_type == 'realtime' else value
        return row_info_dict

    def _get_main_proc_points(self, row_info_dict):
        """
        Get iteration_interval, fp_and_bp and tail points.

        Args:
            row_info_dict (dict): Step trace information.

        Returns:
            list[dict], the list of time points.
        """
        start_point = row_info_dict.get('start_point', 0)
        fp_point = row_info_dict.get('fp_point', 0)
        bp_point = row_info_dict.get('bp_point', 0)
        points_part = [
            self._construct_time_point(
                'iteration_interval', 0, row_info_dict.get('iteration_interval', 0)),
        ]
        # if fp key exist, inference scene
        if 'fp' in row_info_dict.keys():
            points = [
                self._construct_time_point(
                    'fp', fp_point - start_point, row_info_dict.get('fp', 0)),
            ]
        # training scene
        else:
            points = [
                self._construct_time_point(
                    'fp_and_bp', fp_point - start_point, row_info_dict.get('fp_and_bp', 0)),
                self._construct_time_point('tail', bp_point - start_point, row_info_dict.get('tail', 0))
            ]
        points = points_part + points
        return points

    def _get_reduce_time_in_order(self, row_info_dict):
        """
        Get reduce time in order.

        Args:
            row_info_dict (dict): Step trace information.

        Returns:
            dict, sorted reduce information. The reduce info is format like:
            {stream_id: List[Tuple(start_point, end_point, duration, field_name)]}
        """
        reduce_info = {}
        reduce_fields = [field_name for field_name in self.__column__
                         if field_name.startswith('stream_') and not field_name.endswith('point')]
        for reduce_field in reduce_fields:
            reduce_start = row_info_dict.get(reduce_field + '_start_point', 0)
            reduce_end = row_info_dict.get(reduce_field + '_end_point', 0)
            reduce_duration = row_info_dict.get(reduce_field, 0)
            if not (reduce_start and reduce_end and reduce_duration):
                log.info("Reduce event missing value.")
                continue
            cur_stream_id = reduce_field.split('_', 2)[1]
            cur_stream = reduce_info.get(cur_stream_id)
            if not cur_stream:
                cur_stream = []
                reduce_info[cur_stream_id] = cur_stream
            cur_stream.append((reduce_start, reduce_end, reduce_duration, reduce_field))
        for _, reduce_events in reduce_info.items():
            reduce_events.sort(key=lambda elem: elem[1])
        return reduce_info

    def _sort_reduce_by_time(self, row_info_dict):
        """
        Sort reduce info by time.

        Args:
            row_info_dict (dict): Step trace information.

        Returns:
            list, including the all reduce info sorted by start time only.
            [
                [reduce_field, stream_id, reduce_start, reduce_duration],
                [...],
                [...]
            ]
        """
        factor = 1e5  # convert time unit from 10ns to 1ms
        reduce_pid = 10000
        reduce_info = []
        reduce_fields = [field_name for field_name in self.__column__
                         if field_name.startswith('stream_') and not field_name.endswith('point')]
        for reduce_field in reduce_fields:
            reduce_start = row_info_dict.get(reduce_field + '_start_point')
            reduce_start = reduce_start / factor \
                if reduce_start else 0
            reduce_duration = row_info_dict.get(reduce_field)
            reduce_duration = reduce_duration / factor if reduce_duration else 0
            if not (reduce_start and reduce_duration):
                log.info("Reduce event missing value.")
                continue
            cur_stream_id = reduce_field.split('_', 2)[1]
            reduce_meta = [reduce_field, int(cur_stream_id), reduce_start,
                           reduce_duration, reduce_pid]
            reduce_info.append(reduce_meta)

        return reduce_info

    def _construct_reduce_lines(self, row_info_dict):
        """
        Construct first line in detailed graph.

        Args:
            row_info_dict (dict): Step trace information.

        Returns:
            list, list of reduce information of each stream. Each item is a list of time points.
        """
        reduce_lines = []
        start_point = row_info_dict.get('start_point', 0)
        fp_point = row_info_dict.get('fp_point', 0)
        end_point = row_info_dict.get('end_point', 0)
        reduce_info = self._get_reduce_time_in_order(row_info_dict)
        # construct time point for each line
        for _, reduce_events in reduce_info.items():
            current_line = self._construct_reduce_line(
                start_point, end_point, fp_point, reduce_events)
            reduce_lines.append(current_line)

        return reduce_lines

    def _construct_reduce_line(self, start_point, end_point, fp_point, reduce_events):
        """
        Construct list of time points for reduce line.

        Args:
            start_point (int): The start point of current step.
            end_point (int): The end point of current step.
            fp_point (int): The fp point of current step.
            reduce_events (list[Tuple]): The reduce information of current step. Each item
                contains the start, end duration and name of one reduce event.

        Returns:
            list[dict], list of time points.
        """
        current_line = []
        previous_start = fp_point
        for start, end, duration, field_name in reduce_events:
            current_line.extend([
                self._construct_time_point(
                    '', previous_start - start_point, start - previous_start),
                self._construct_time_point(
                    field_name, start - start_point, duration)
            ])
            previous_start = end
        current_line.append(self._construct_time_point(
            '', previous_start - start_point, end_point - previous_start))
        return current_line

    def _get_proc_details(self, proc_name, step_id=None, time_type='realtime'):
        """
        Get step trace info for selected step and save the result.

        Args:
            proc_name (str): The selected field name.
            step_id (int): The selected step_id. If not given, it means all steps is required.
                If the value is 0, it means average info for all steps except the first is
                required. Default: None.
            time_type (str): The value type. `systime` keeps the original value.
                `realtime` transforms the value in millisecond. Default: `realtime`.
        """
        if proc_name is None:
            log.error('`proc_name` is required for query.')
            raise ProfilerParamValueErrorException('`proc_name` is required for query.')
        if step_id is None:
            rows_info = self._data[:-1]
        else:
            rows_info = [self._data[step_id - 1]]

        proc_info = [get_field_value(row_info, proc_name, self.__column__, time_type)
                     for row_info in rows_info]
        self._result['info'] = {proc_name: proc_info}

    def _validate_filter_condition(self, filter_condition):
        """Validate step trace filter_condition."""
        mode = filter_condition.get('mode', 'step')
        self._validate_str_param(mode, ['step', 'proc'], 'mode')

        step_id = filter_condition.get('step_id')
        self._validate_step_id(step_id)

        proc_name = filter_condition.get('proc_name')
        self._validate_str_param(proc_name, self.__column__, 'proc_name')

        time_type = filter_condition.get('time_type', 'realtime')
        self._validate_str_param(time_type, ['realtime', 'systime'], 'time_type')

    def _validate_step_id(self, step_id):
        """Validate step_id."""
        if step_id is None or isinstance(step_id, int) and 0 <= step_id <= self._size:
            return
        log.error("Invalid step_id in request. step_id should be in [0, %d].", self._size)
        raise StepNumNotSupportedException([0, self._size])

    @staticmethod
    def _validate_str_param(proc_name, accept_param, error_name=''):
        """Validate proc_name."""
        if proc_name is None or isinstance(proc_name, str) and proc_name in accept_param:
            return
        log.error("Invalid param %s in request. Acceptable value is %s.", error_name, accept_param)
        raise ProfilerParamValueErrorException(f"Invalid {error_name}.")



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""This module provides python APIs to get lineage summary from summary log."""
import struct
from collections import namedtuple
from enum import Enum

from google.protobuf.json_format import MessageToDict
from google.protobuf.message import DecodeError

from mindinsight.datavisual.proto_files.mindinsight_lineage_pb2 import LineageEvent
from mindinsight.lineagemgr.common.exceptions.exceptions import MindInsightException, \
    LineageSummaryAnalyzeException, LineageFileHandlerReadError
from mindinsight.lineagemgr.common.log import logger as log
from mindinsight.lineagemgr.common.validator.validate_path import safe_normalize_path
from mindinsight.lineagemgr.summary.file_handler import FileHandler

LineageInfo = namedtuple('LineageInfo', ['train_lineage', 'eval_lineage', 'dataset_graph'])


class SummaryTag(Enum):
    """The tag value of lineage fields."""

    # the value is `field_number << 3 | wire_type`
    WALL_TIME = 'wall_time'
    STEP = 'step'
    VERSION = 'version'
    GRAPH = 'graph'
    SUMMARY = 'summary'
    TRAIN_LINEAGE = 'train_lineage'
    EVAL_LINEAGE = 'evaluation_lineage'
    DATASET_GRAPH = 'dataset_graph'


class SummaryAnalyzer:
    """
    Summary log Analyzer.

    Args:
        file_path (str): The path of summary log.

    Raises:
        LineageSummaryAnalyzeException: Raise when read files failed.
    """
    HEADER_SIZE = 8
    HEADER_CRC_SIZE = 4
    BODY_CRC_SIZE = 4

    def __init__(self, file_path):
        self.file_handler = FileHandler(file_path)

    def load_events(self):
        """
        Load events in summary log.

        Returns:
            generator, the event generator.
        """
        while self._has_next():
            yield self._read_event()

    def _has_next(self):
        """
        Check if the file has reached the end.

        Returns:
            bool, whether the file has reached the end.
        """
        current_offset = self.file_handler.tell()
        if current_offset < self.file_handler.size:
            return True

        return False

    def _read_event(self):
        """
        Read event.

        Returns:
            LineageEvent, the event body.
        """
        body_size = self._read_header()
        body_str = self._read_body(body_size)
        event = LineageEvent().FromString(body_str)
        return event

    def _read_header(self):
        """
        Read header information.

        Returns:
            int, the length of event body.
        """
        header_str = self.file_handler.read(self.HEADER_SIZE)
        self.file_handler.read(self.HEADER_CRC_SIZE)
        body_len = struct.unpack("<Q", header_str)[0]

        return body_len

    def _read_body(self, body_size):
        """
        Read event body information.

        Args:
            body_size (int): The size of event body.

        Returns:
            bytes, the event body in bytes.
        """
        body_str = self.file_handler.read(body_size)
        self.file_handler.read(self.BODY_CRC_SIZE)

        return body_str


class LineageSummaryAnalyzer(SummaryAnalyzer):
    """
    Summary log analyzer for lineage information.

    Args:
        file_path (str): The path of summary log.

    Raises:
        LineageSummaryAnalyzeException: If failed to get lineage information.
    """

    def __init__(self, file_path):
        file_path = safe_normalize_path(file_path, 'lineage_summary_path', None)
        super(LineageSummaryAnalyzer, self).__init__(file_path)

    def get_latest_info(self):
        """
        Get latest lineage info in summary log file.

        Returns:
            LineageInfo, the lineage summary information.
        """
        lineage_events = {
            SummaryTag.TRAIN_LINEAGE: None,
            SummaryTag.EVAL_LINEAGE: None,
            SummaryTag.DATASET_GRAPH: None
        }
        for event in self.load_events():
            for tag, _ in lineage_events.items():
                if event.HasField(tag.value):
                    lineage_events[tag] = event
                    break

        lineage_info = LineageInfo(
            train_lineage=lineage_events.get(SummaryTag.TRAIN_LINEAGE),
            eval_lineage=lineage_events.get(SummaryTag.EVAL_LINEAGE),
            dataset_graph=lineage_events.get(SummaryTag.DATASET_GRAPH)
        )

        return lineage_info

    @classmethod
    def get_summary_infos(cls, file_path):
        """
        Get lineage summary information from summary log file.

        Args:
            file_path (str): The file path of summary log.

        Returns:
            LineageInfo, the lineage summary information.

        Raises:
            LineageSummaryAnalyzeException: If failed to get lineage information.
        """
        analyzer = cls(file_path)
        err_msg = "Can not analyze lineage info, file path is %s. Detail: %s"
        try:
            lineage_info = analyzer.get_latest_info()
        except (MindInsightException, IOError, DecodeError, LineageFileHandlerReadError) as err:
            log.debug(err_msg, file_path, str(err))
            raise LineageSummaryAnalyzeException(str(err))
        except Exception as err:
            log.debug(err_msg, file_path, str(err))
            raise LineageSummaryAnalyzeException(str(err))

        return lineage_info

    @staticmethod
    def get_user_defined_info(file_path):
        """
        Get user defined info.
        Args:
            file_path (str): The file path of summary log.

        Returns:
            list, the list of dict format user defined information
                which converted from proto message.
        """
        all_user_message = []
        summary_analyzer = SummaryAnalyzer(file_path)

        for event in summary_analyzer.load_events():
            if event.HasField("user_defined_info"):
                user_defined_info = MessageToDict(
                    event,
                    preserving_proto_field_name=True
                ).get("user_defined_info")
                user_dict = LineageSummaryAnalyzer._get_dict_from_proto(user_defined_info)
                all_user_message.append(user_dict)

        return all_user_message

    @staticmethod
    def _get_dict_from_proto(user_defined_info):
        """
        Convert the proto message UserDefinedInfo to its dict format.

        Args:
            user_defined_info (UserDefinedInfo): The proto message of user defined info.

        Returns:
            dict, the converted dict.
        """
        user_dict = dict()
        proto_dict = user_defined_info.get("user_info")
        for proto_item in proto_dict:
            if proto_item and isinstance(proto_item, dict):
                key, value = list(list(proto_item.values())[0].items())[0]
                if isinstance(value, dict):
                    user_dict[key] = LineageSummaryAnalyzer._get_dict_from_proto(value)
                else:
                    user_dict[key] = value

        return user_dict



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
MindSpore Debugger is a debugging tool for training in Graph Mode. It can be applied to visualize
and analyze the intermediate computation results of the computational graph.
In Graph Mode training, the computation results of intermediate nodes in the computational graph can not be acquired
conveniently, which makes it difficult for users to do the debugging.

By applying MindSpore Debugger, users can:Visualize the computational graph on the UI and analyze the output
of the graph node.Set watchpoints to monitor training exceptions (for example, tensor overflow) and trace error causes.
Visualize and analyze the change of parameters, such as weights.Visualize the nodes and code mapping relationship.

Debugger API is a python API interface provided for offline debugger. You need to save dump data before using it.
For the method of saving dump data, `refer to Using Dump in the Graph Mode
<https://www.mindspore.cn/tutorials/experts/en/master/debug/dump.html>`_ .
"""

from mindinsight.debugger.api.conditions import \
    Watchpoint, WatchpointHit, TensorTooLargeCondition, TensorUnchangedCondition, TensorAllZeroCondition, \
    TensorOverflowCondition, OperatorOverflowCondition, TensorRangeCondition, TensorTooSmallCondition, \
    TensorChangeBelowThresholdCondition, TensorChangeAboveThresholdCondition, ConditionBase
from mindinsight.debugger.api.debugger_tensor import DebuggerTensor
from mindinsight.debugger.api.dump_analyzer import DumpAnalyzer
from mindinsight.debugger.api.node import Node


__all__ = ["DumpAnalyzer", "Node", "DebuggerTensor", "Watchpoint",
           "WatchpointHit",
           "TensorTooLargeCondition",
           "TensorTooSmallCondition",
           "TensorRangeCondition",
           "TensorOverflowCondition",
           "OperatorOverflowCondition",
           "TensorAllZeroCondition",
           "TensorUnchangedCondition",
           "TensorChangeBelowThresholdCondition",
           "TensorChangeAboveThresholdCondition",
           "ConditionBase"
           ]



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""This file provides path resolution."""
from mindinsight.datavisual.data_transform.summary_watcher import SummaryWatcher
from mindinsight.lineagemgr.common.log import logger
from mindinsight.lineagemgr.common.utils import get_timestamp
from mindinsight.utils.exceptions import MindInsightException


class SummaryPathParser:
    """
    Summary path parser.

    This class is a utility class, users can use it to parse summary dir,
    parse summary log path, get the latest lineage summary log, etc.
    """
    LINEAGE_SUMMARY_SUFFIX = '_lineage'
    _LINEAGE_SUMMARY_SUFFIX_LEN = len(LINEAGE_SUMMARY_SUFFIX)

    @staticmethod
    def get_lineage_summaries(summary_dir, is_sorted=False, reverse=True):
        """
        Get lineage summaries according to summary dir.

        Args:
            summary_dir (str): Summary dir.
            is_sorted (bool): If it is True, files will be sorted.
            reverse (bool): If it is True, sort by timestamp increments and filename decrement.

        Returns:
            list, if the lineage summary log exist, return the file names, else return [].
        """
        try:
            summary_watcher = SummaryWatcher()
            summaries = summary_watcher.list_summaries(summary_base_dir=summary_dir)
        except MindInsightException as err:
            logger.warning(str(err))
            return []
        summary_files = [summary.get('file_name') for summary in summaries]
        lineage_files_name = list(filter(
            lambda filename: (filename.endswith(SummaryPathParser.LINEAGE_SUMMARY_SUFFIX)), summary_files))
        if is_sorted:
            lineage_files_name = SummaryPathParser._sorted_summary_files(lineage_files_name, reverse)

        return lineage_files_name

    @staticmethod
    def _sorted_summary_files(summary_files, reverse):
        """Sort by timestamp increments and filename decrement."""
        sorted_files = sorted(summary_files,
                              key=lambda filename: (-get_timestamp(filename), filename),
                              reverse=reverse)
        return sorted_files



# Copyright 2019-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Backend interface module.

This module provides the interfaces to train processors functions.
"""
from flask import Blueprint
from flask import request
from flask import jsonify

from mindinsight.conf import settings
from mindinsight.datavisual.utils.tools import if_nan_inf_to_none
from mindinsight.datavisual.processors.histogram_processor import HistogramProcessor
from mindinsight.datavisual.processors.tensor_processor import TensorProcessor
from mindinsight.datavisual.processors.images_processor import ImageProcessor
from mindinsight.datavisual.processors.scalars_processor import ScalarsProcessor
from mindinsight.datavisual.processors.graph_processor import GraphProcessor
from mindinsight.datavisual.data_transform.data_manager import DATA_MANAGER
from mindinsight.datavisual.processors.landscape_processor import LandscapeProcessor
from mindinsight.utils.exceptions import ParamValueError


VALID_MODE = {"normal", "optimize"}
BLUEPRINT = Blueprint("train_visual", __name__, url_prefix=settings.URL_PATH_PREFIX+settings.API_PREFIX)


@BLUEPRINT.route("/datavisual/image/metadata", methods=["GET"])
def image_metadata():
    """
    Interface to fetch metadata about the images for the particular run,tag, and zero-indexed sample.

    Returns:
        Response, which contains a list in JSON containing image events, each
            one of which is an object containing items wall_time, step, width,
            height, and query.
    """
    tag = request.args.get("tag")
    train_id = request.args.get('train_id')

    processor = ImageProcessor(DATA_MANAGER)
    response = processor.get_metadata_list(train_id, tag)
    return jsonify(response)


@BLUEPRINT.route("/datavisual/image/single-image", methods=["GET"])
def single_image():
    """
    Interface to fetch raw image data for a particular image.

    Returns:
        Response, which contains a byte string of image.
    """
    tag = request.args.get("tag")
    step = request.args.get("step")
    train_id = request.args.get('train_id')

    processor = ImageProcessor(DATA_MANAGER)
    img_data = processor.get_single_image(train_id, tag, step)
    return img_data


@BLUEPRINT.route("/datavisual/scalar/metadata", methods=["GET"])
def scalar_metadata():
    """
    Interface to fetch metadata about the scalars for the particular run and tag.

    Returns:
        Response, which contains a list in JSON containing scalar events, each
            one of which is an object containing items' wall_time, step and value.
    """
    tag = request.args.get("tag")
    train_id = request.args.get('train_id')

    processor = ScalarsProcessor(DATA_MANAGER)
    response = processor.get_metadata_list(train_id, tag)

    metadatas = response['metadatas']
    for metadata in metadatas:
        value = metadata.get("value")
        metadata["value"] = if_nan_inf_to_none('scalar_value', value)

    return jsonify(response)


@BLUEPRINT.route("/datavisual/graphs/nodes", methods=["GET"])
def graph_nodes():
    """
    Interface to get graph nodes.

    Returns:
        Response, which contains a JSON object.

    """
    name = request.args.get('name', default=None)
    tag = request.args.get("tag", default=None)
    mode = request.args.get("mode", default="normal")
    train_id = request.args.get('train_id')
    if mode not in VALID_MODE:
        raise ParamValueError("Invalid mode")

    graph_process = GraphProcessor(train_id, DATA_MANAGER, tag, mode)
    response = graph_process.list_nodes(scope=name)
    return jsonify(response)


@BLUEPRINT.route("/datavisual/graphs/nodes/names", methods=["GET"])
def graph_node_names():
    """
    Interface to query node names.

    Returns:
        Response, which contains a JSON object.
    """
    search_content = request.args.get("search")
    offset = request.args.get("offset", default=0)
    limit = request.args.get("limit", default=100)
    tag = request.args.get("tag", default=None)
    mode = request.args.get("mode", default="normal")
    train_id = request.args.get('train_id')
    if mode not in VALID_MODE:
        raise ParamValueError("Invalid mode")

    graph_process = GraphProcessor(train_id, DATA_MANAGER, tag, mode)
    resp = graph_process.search_node_names(search_content, offset, limit)
    return jsonify(resp)


@BLUEPRINT.route("/datavisual/graphs/nodes/all", methods=["GET"])
def graph_all_nodes():
    """
    Interface to get all nodes detail.

    Returns:
        Response, which contains a JSON object.
    """
    name = request.args.get("name", default=None)
    tag = request.args.get("tag", default=None)
    mode = request.args.get("mode", default="normal")
    train_id = request.args.get('train_id')
    if mode not in VALID_MODE:
        raise ParamValueError("Invalid mode")

    graph_process = GraphProcessor(train_id, DATA_MANAGER, tag, mode)
    resp = graph_process.list_all_nodes_info(scope=name)
    return jsonify(resp)


@BLUEPRINT.route("/datavisual/graphs/single-node", methods=["GET"])
def graph_search_single_node():
    """
    Interface to search single node.

    Returns:
         Response, which contains a JSON object.
    """
    name = request.args.get("name")
    tag = request.args.get("tag", default=None)
    mode = request.args.get("mode", default="normal")
    train_id = request.args.get('train_id')
    if mode not in VALID_MODE:
        raise ParamValueError("Invalid mode")

    graph_process = GraphProcessor(train_id, DATA_MANAGER, tag, mode)
    resp = graph_process.search_single_node(name)
    return jsonify(resp)


@BLUEPRINT.route("/datavisual/histograms", methods=["GET"])
def histogram():
    """
    Interface to obtain histogram data.

    Returns:
        Response, which contains a JSON object.
    """
    tag = request.args.get("tag", default=None)
    train_id = request.args.get('train_id')

    processor = HistogramProcessor(DATA_MANAGER)
    response = processor.get_histograms(train_id, tag)
    return jsonify(response)


@BLUEPRINT.route("/datavisual/scalars", methods=["GET"])
def get_scalars():
    """Get scalar data for given train_ids and tags."""
    train_ids = request.args.getlist('train_id')
    tags = request.args.getlist('tag')

    processor = ScalarsProcessor(DATA_MANAGER)
    scalars = processor.get_scalars(train_ids, tags)
    return jsonify({'scalars': scalars})


@BLUEPRINT.route("/datavisual/tensors", methods=["GET"])
def get_tensors():
    """
    Interface to obtain tensor data.

    Returns:
        Response, which contains a JSON object.
    """
    train_ids = request.args.getlist('train_id')
    tags = request.args.getlist('tag')
    step = request.args.get("step", default=None)
    dims = request.args.get("dims", default=None)
    detail = request.args.get("detail", default=None)

    processor = TensorProcessor(DATA_MANAGER)
    response = processor.get_tensors(train_ids, tags, step, dims, detail)
    return jsonify(response)


def init_module(app):
    """
    Init module entry.

    Args:
        app (Flask): The application obj.
    """
    app.register_blueprint(BLUEPRINT)


@BLUEPRINT.route("/datavisual/landscape/intervals", methods=["GET"])
def list_intervals():
    """
    Get the intervals.

    Returns:
        Response, which contains a JSON object.
    """
    train_id = request.args.get('train_id')
    processor = LandscapeProcessor(DATA_MANAGER)
    response = processor.list_intervals(train_id)
    return jsonify(response)


@BLUEPRINT.route("/datavisual/landscape/landscapes", methods=["GET"])
def list_landscapes():
    """
    Get the loss landscape data.

    Returns:
        Response, which contains a JSON object.
    """
    train_ids = request.args.getlist('train_id')
    landscape_type = request.args.get('type', 'interval')
    interval_id = request.args.get('interval_id', 0)

    processor = LandscapeProcessor(DATA_MANAGER)
    response = processor.list_landscapes(train_ids, landscape_type, interval_id)
    return jsonify(response)



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Import the streams handlers."""
from .event_handler import EventHandler
from .metadata_handler import MetadataHandler
from .graph_handler import GraphHandler, MultiCardGraphHandler
from .tensor_handler import TensorHandler, MultiCardTensorHandler
from .watchpoint_handler import WatchpointHandler, WatchpointHitHandler, MultiCardWatchpointHitHandler

__all__ = ['EventHandler', 'MetadataHandler', 'GraphHandler', 'TensorHandler', 'WatchpointHitHandler',
           'MultiCardGraphHandler', 'MultiCardTensorHandler',
           'WatchpointHandler', 'MultiCardWatchpointHitHandler']



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Utility functions for hierarchcial occlusion image generation."""

import re

import PIL
from PIL import ImageDraw, ImageEnhance, ImageFilter

MASK_GAUSSIAN_RE = r'^gaussian:(\d+)$'


class EditStep:
    """
    Class that represents an edit step.

    Args:
        layer (int): Layer index.
        x (int): Left pixel coordinate.
        y (int): Top pixel coordinate.
        width (int): Width in pixels.
        height (int): Height in pixels.
    """
    def __init__(self,
                 layer: int,
                 x: int,
                 y: int,
                 width: int,
                 height: int):
        self.layer = layer
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def to_coord_box(self):
        """
        Convert to pixel coordinate box.

        Returns:
            tuple[int, int, int, int], tuple of left, top, right, bottom pixel coordinate.
        """
        return self.x, self.y, self.x + self.width, self.y + self.height


def pil_apply_edit_steps(image, mask, edit_steps, by_masking=False, inplace=False):
    """
    Apply edit steps on a PIL image.

    Args:
        image (PIL.Image): The input image in RGB mode.
        mask (Union[str, int, tuple[int, int, int], PIL.Image.Image]): The mask to apply on the image, could be string
            e.g. 'gaussian:9', a single, grey scale intensity [0, 255], an RBG tuple or a PIL Image object.
        edit_steps (list[EditStep]): Edit steps to be drawn.
        by_masking (bool): Whether to use masking method. Default: False.
        inplace (bool): True to draw on the input image, otherwise draw on a cloned image.

    Returns:
        PIL.Image, the result image.
    """
    if isinstance(mask, str):
        mask = pil_compile_str_mask(mask, image)
    if by_masking:
        return _pil_apply_edit_steps_mask(image, mask, edit_steps, inplace)
    return _pil_apply_edit_steps_unmask(image, mask, edit_steps, inplace)


def pil_compile_str_mask(mask, image):
    """Concert string mask to PIL Image."""
    match = re.match(MASK_GAUSSIAN_RE, mask)
    if match:
        radius = int(match.group(1))
        if radius > 0:
            image_filter = ImageFilter.GaussianBlur(radius=radius)
            mask_image = image.filter(image_filter)
            mask_image = ImageEnhance.Brightness(mask_image).enhance(0.7)
            mask_image = ImageEnhance.Color(mask_image).enhance(0.0)
            return mask_image
    raise ValueError(f"Invalid string mask: '{mask}'.")


def _pil_apply_edit_steps_unmask(image, mask, edit_steps, inplace=False):
    """
    Apply edit steps from unmasking method on a PIL image.

    Args:
        image (PIL.Image): The input image.
        mask (Union[int, tuple[int, int, int], PIL.Image]): The mask to apply on the image, could be a single grey
            scale intensity [0, 255], an RBG tuple or a PIL Image.
        edit_steps (list[EditStep]): Edit steps to be drawn.
        inplace (bool): True to draw on the input image, otherwise draw on a cloned image.

    Returns:
        PIL.Image, the result image.
    """
    if isinstance(mask, PIL.Image.Image):
        if inplace:
            bg = mask
        else:
            bg = mask.copy()
    else:
        if inplace:
            raise ValueError('Argument inplace cannot be True when mask is not a PIL Image.')
        if isinstance(mask, int):
            mask = (mask, mask, mask)
        bg = PIL.Image.new(mode="RGB", size=image.size, color=mask)

    for step in edit_steps:
        box = step.to_coord_box()
        cropped = image.crop(box)
        bg.paste(cropped, box=box)
    return bg


def _pil_apply_edit_steps_mask(image, mask, edit_steps, inplace=False):
    """
    Apply edit steps from unmasking method on a PIL image.

    Args:
        image (PIL.Image): The input image.
        mask (Union[int, tuple[int, int, int], PIL.Image]): The mask to apply on the image, could be a single grey
            scale intensity [0, 255], an RBG tuple or a PIL Image.
        edit_steps (list[EditStep]): Edit steps to be drawn.
        inplace (bool): True to draw on the input image, otherwise draw on a cloned image.

    Returns:
        PIL.Image, the result image.
    """
    if not inplace:
        image = image.copy()

    if isinstance(mask, PIL.Image.Image):
        for step in edit_steps:
            box = step.to_coord_box()
            cropped = mask.crop(box)
            image.paste(cropped, box=box)
    else:
        if isinstance(mask, int):
            mask = (mask, mask, mask)
        draw = ImageDraw.Draw(image)
        for step in edit_steps:
            draw.rectangle(step.to_coord_box(), fill=mask)
    return image



# Copyright 2019-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Stop mindinsight service."""

import os
import sys
import argparse
import signal
import getpass

import psutil

from mindinsight.conf import settings
from mindinsight.utils.command import BaseCommand
from mindinsight.utils.hook import HookUtils
from mindinsight.modelarts.utils.utils import get_notebook_registry


class PortAction(argparse.Action):
    """Port action class definition."""

    MIN_PORT = 1
    MAX_PORT = 65535

    def __call__(self, parser, namespace, values, option_string=None):
        """
        Inherited __call__ method from argparse.Action.

        Args:
            parser (ArgumentParser): Passed-in argument parser.
            namespace (Namespace): Namespace object to hold arguments.
            values (object): Argument values with type depending on argument definition.
            option_string (str): Optional string for specific argument name. Default: None.
        """
        port = values
        if not self.MIN_PORT <= port <= self.MAX_PORT:
            parser.error(f'{option_string} should be chosen from {self.MIN_PORT} to {self.MAX_PORT}')

        setattr(namespace, self.dest, port)


class Command(BaseCommand):
    """Stop command."""
    name = 'stop'
    description = 'stop mindinsight service'

    cmd_regex = 'mindinsight.backend.application:APP'

    def add_arguments(self, parser):
        """
        Add arguments to parser.

        Args:
            parser (ArgumentParser): Specify parser to which arguments are added.
        """
        parser.add_argument(
            '--port',
            type=int,
            action=PortAction,
            help="""
                Custom port ranging from %s to %s. Default value is %s
            """ % (PortAction.MIN_PORT, PortAction.MAX_PORT, settings.PORT))

    def update_settings(self, args):
        """
        Update settings.

        Args:
            args (Namespace): parsed arguments to hold customized parameters.
        """
        if args.port is None:
            args.port = settings.PORT

        pid, workspace, log_base_dir = self.get_process(args.port)
        setattr(args, 'pid', pid)
        setattr(args, 'log_base_dir', log_base_dir)

        os.environ['MINDINSIGHT_PORT'] = str(args.port)
        os.environ['MINDINSIGHT_WORKSPACE'] = workspace
        settings.refresh()

    def run(self, args):
        """
        Run to stop.

        Args:
            args (Namespace): Parsed arguments to hold customized parameters.
        """
        port, pid = args.port, args.pid
        notebook_register = get_notebook_registry()
        if notebook_register:
            notebook_register.stop(port)

        if not pid:
            msg = f'No mindinsight service started by current user found for port {port}'
            self.console.error(msg)
            sys.exit(1)

        self.logfile.info('Stop mindinsight with port %s and pid %s.', port, pid)

        process = psutil.Process(pid)
        processes = process.children(recursive=True)
        processes.append(process)
        try:
            self._send_signal(process, signal.SIGINT)
            # Wait 3 seconds, if not terminate, kill the worker process.
            exit_timeout_seconds = 3
            _, alive = psutil.wait_procs(processes, exit_timeout_seconds)
            for alive_process in alive:
                self.logfile.info("Stop process %d because timeout.", alive_process.pid)
                self._send_signal(alive_process, signal.SIGKILL)
        except psutil.Error as ex:
            self.logfile.error("Stop process %d failed. Detail: %s.", pid, str(ex))

        for hook in HookUtils.instance().hooks():
            hook.on_shutdown(self.logfile)

        self.console.info('Stop mindinsight service successfully')

    def _send_signal(self, process, proc_signal):
        try:
            process.send_signal(proc_signal)
        except psutil.NoSuchProcess:
            pass

    def get_process(self, port):
        """
        Get mindinsight process

        Args:
            port (int): Specified port for mindinsight process.

        Returns:
            tuple, return mindinsight process pid and workspace.
        """
        pid, workspace = 0, settings.WORKSPACE
        user = getpass.getuser()
        connections = psutil.net_connections()
        log_base_dir = ''
        for connection in connections:
            if connection.status != 'LISTEN':
                continue
            if connection.laddr.port != port:
                continue

            try:
                process = psutil.Process(connection.pid)
            except psutil.NoSuchProcess:
                continue

            cmds = process.cmdline()
            if ' '.join(cmds).find(self.cmd_regex) == -1:
                continue
            if user != process.username():
                continue

            gunicorn_master_process = process

            # The gunicorn master process might have grand children (eg forked by process pool).
            while True:
                parent_process = gunicorn_master_process.parent()
                if parent_process is None or parent_process.pid == 1:
                    break
                parent_cmd = parent_process.cmdline()
                if ' '.join(parent_cmd).find(self.cmd_regex) == -1:
                    break
                gunicorn_master_process = parent_process

            pid = gunicorn_master_process.pid

            access_log_path = os.path.join('gunicorn', 'access.{}.log'.format(port))
            for open_file in process.open_files():
                if open_file.path.endswith(access_log_path):
                    log_base_dir = open_file.path[:-len(access_log_path)]
                    workspace = os.path.realpath(os.path.join(log_base_dir, os.pardir))
                    break
            break

        return pid, workspace, log_base_dir



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Modelarts notebook module."""
import json
import os
import random
import argparse
import shlex
import logging
import uuid
import subprocess

import mindinsight
from mindinsight.scripts.start import Command
from mindinsight.utils.log import setup_logger
from mindinsight.modelarts.utils.utils import get_notebook_registry


logger = setup_logger("notebook", "notebook", console=True, level=logging.INFO)

_OK = 0
_NOK = 1
_REUSE = 2

try:
    import html

    html_escape = html.escape
    del html
except ImportError:
    import cgi

    html_escape = cgi.escape
    del cgi


# Return values for `_get_context` (see that function's docs for details).
_CONTEXT_IPYTHON = "_CONTEXT_IPYTHON"
_CONTEXT_NONE = "_CONTEXT_NONE"


def _get_context():
    """
    Determine the most specific context that we're in.

    Returns:
      str: in an IPython notebook context (e.g., from running `jupyter notebook` at the command line).
      str: Otherwise (e.g., by running a Python script at the
        command-line or using the `ipython` interactive shell).
    """
    try:
        import IPython
    except ImportError:
        pass
    else:
        if IPython.get_ipython() is not None:
            # We'll assume that we're in a notebook context.
            return _CONTEXT_IPYTHON

    # In an IPython command line shell or Jupyter notebook, we can
    # directly query whether we're in a notebook context.
    try:
        import IPython
    except ImportError:
        pass
    else:
        ipython = IPython.get_ipython()
        if ipython is not None and ipython.has_trait("kernel"):
            return _CONTEXT_IPYTHON

    # Otherwise, we're not in a known notebook context.
    return _CONTEXT_NONE


def notebook_load_ipython_extension(ipython):
    """
    Load the MindInsight notebook extension.

    Intended to be called from `%load_ext mindinsight`. Do not invoke this directly.

    Args:
      ipython(ipykernel.zmqshell.ZMQInteractiveShell): An `IPython.InteractiveShell` instance.
    """
    _register_magics(ipython)


def _register_magics(ipython):
    """
    Register IPython line/cell magics.

    Args:
      ipython(ipykernel.zmqshell.ZMQInteractiveShell): An `InteractiveShell` instance.
    """
    ipython.register_magic_function(
        _start_magic, magic_kind="line", magic_name="mindinsight",
    )


def _start_magic(line):
    """Implementation of the `%mindinsight` line magic."""
    if os.environ.get('BASE_URL', None) is None:
        return print("MindInsight is not supported in the current environment!")
        
    return start(line)

def start(args_string):
    """
    Launch and display a MindInsight instance as if at the command line.

    Args:
      args_string(str): Command-line arguments to MindInsight, to be
        interpreted by `shlex.split`: e.g., "--summary-base-dir ./logs --port 8080".
        Shell metacharacters are not supported: e.g., "--summary-base-dir 2>&1" will
        point the summary-base-dir at the literal directory named "2>&1".
    """
    parsed_args = shlex.split(args_string, comments=True, posix=True)

    parser = argparse.ArgumentParser(
        prog='mindinsight',
        description='MindInsight CLI entry point (version: {})'.format(mindinsight.__version__),
        allow_abbrev=False)

    notebook_command = Command()
    notebook_command.add_arguments(parser=parser)
    args = parser.parse_args(parsed_args)

    context = _get_context()
    try:
        import IPython
        import IPython.display
    except ImportError:
        IPython = None

    if context == _CONTEXT_NONE:
        handle = None
        logger.info("Launching MindInsight...")
    else:
        handle = IPython.display.display(
            IPython.display.Pretty("Launching MindInsight in Notebook..."), display_id=True,
        )

    input_path = ''
    port = 8080
    for key, value in args.__dict__.items():
        if key == 'summary_base_dir' and value:
            input_path = value
        if key == 'port' and value:
            port = int(value)

    notebook_register = get_notebook_registry()
    if notebook_register:
        args_string = notebook_register.start(args_string, port=port, input_path=input_path)

    cmd = 'mindinsight start {}'.format(args_string)
    process = subprocess.Popen(
        shlex.split(cmd),
        shell=False,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT
    )
    process.wait()
    mindinsight_stdout = process.stdout.read().decode('utf-8')
    if process.returncode == _OK or process.returncode == _REUSE:
        if process.returncode == _OK:
            logger.info(f'mindinsight start on port {port}')
        if process.returncode == _REUSE:
            logger.info("{}".format(mindinsight_stdout))
        display(port=port)
    else:
        logger.info("{}".format(mindinsight_stdout))


def display(port=None, height=None):
    """
    Display a MindInsight instance already running on this machine.

    Args:
      port(Union[None, int]): The port on which the MindInsight server is listening, as an
        `int`, or `None` to automatically select the most recently
        launched MindInsight.
      height(Union[None, int]): The height of the frame into which to render the MindInsight
        UI, as an `int` number of pixels, or `None` to use a default value
        (currently 800).
    """
    _display(port=port, height=height, display_handle=None)


def _display(port=None, height=None, display_handle=None):
    """
    Internal version of `display`.

    Args:
      port(Union[None, int]): As with `display`.
      height(Union[None, int]): As with `display`.
      display_handle(NoneType): If not None, an IPython display handle into which to
        render MindInsight.
    """
    if height is None:
        height = 800

    if port is None:
        port = 8080

    fn = {
        _CONTEXT_IPYTHON: _display_ipython,
        _CONTEXT_NONE: _display_cli,
    }[_get_context()]
    return fn(port=port, height=height, display_handle=display_handle)


def _display_ipython(port, height, display_handle):
    import IPython.display

    frame_id = "mindinsight-frame-{:08x}".format(random.getrandbits(64))
    shell = """
      <iframe id="%HTML_ID%" width="100%" height="%HEIGHT%" frameborder="0">
      </iframe>
      <script>
        (function() {
          const frame = document.getElementById(%JSON_ID%);
          const url = new URL(%URL%, window.location);
          const port = %PORT%;
          if (port) {
            url.port = port;
          }
          frame.src = url;
        })();
      </script>
    """
    # proxy_url should be constructed from ``{base_url}/proxy/{port}``
    base_url = os.environ.get('BASE_URL') if 'BASE_URL' in os.environ and os.environ['BASE_URL'] else ''
    default_proxy_url = base_url + '/proxy/%PORT%/'
    proxy_url = os.environ.get('PROXY_URL') if 'PROXY_URL' in os.environ and os.environ[
        'PROXY_URL'] else default_proxy_url

    if proxy_url is not None:
        # Allow %PORT% in $PROXY_URL
        proxy_url = proxy_url.replace("%PORT%", "%d" % port)
        replacements = [
            ("%HTML_ID%", html_escape(frame_id, quote=True)),
            ("%JSON_ID%", json.dumps(frame_id)),
            ("%HEIGHT%", "%d" % height),
            ("%PORT%", "0"),
            ("%URL%", json.dumps(proxy_url)),
        ]
    else:
        replacements = [
            ("%HTML_ID%", html_escape(frame_id, quote=True)),
            ("%JSON_ID%", json.dumps(frame_id)),
            ("%HEIGHT%", "%d" % height),
            ("%PORT%", "%d" % port),
            ("%URL%", json.dumps("/")),
        ]

    for (k, v) in replacements:
        shell = shell.replace(k, v)
    iframe = IPython.display.HTML(shell)
    if display_handle:
        display_handle.update(iframe)
    else:
        IPython.display.display(iframe)


def _display_cli(port, height, display_handle):
    del height  # unused
    del display_handle  # unused
    message = "Please visit http://localhost:%d in a web browser." % port
    logger.info(message)



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""This file is used to define the model lineage python api."""
import numpy as np
import pandas as pd

from mindinsight.lineagemgr.common.exceptions.exceptions import LineageQuerySummaryDataError, \
    LineageQuerierParamException, LineageSearchConditionParamError, LineageParamTypeError, LineageSummaryParseException
from mindinsight.lineagemgr.common.log import logger as log
from mindinsight.lineagemgr.common.validator.model_parameter import SearchModelConditionParameter
from mindinsight.lineagemgr.common.validator.validate import validate_search_model_condition, validate_condition
from mindinsight.lineagemgr.common.validator.validate_path import validate_and_normalize_path
from mindinsight.lineagemgr.lineage_parser import LineageOrganizer
from mindinsight.lineagemgr.querier.querier import Querier
from mindinsight.optimizer.common.enums import ReasonCode
from mindinsight.optimizer.utils.utils import is_simple_numpy_number
from mindinsight.utils.exceptions import MindInsightException

METRIC_PREFIX = "[M]"
USER_DEFINED_PREFIX = "[U]"

USER_DEFINED_INFO_LIMIT = 100


def filter_summary_lineage(data_manager=None, summary_base_dir=None, search_condition=None):
    """
    Filter summary lineage from data_manager or parsing from summaries.

    One of data_manager or summary_base_dir needs to be specified. Support getting
    super_lineage_obj from data_manager or parsing summaries by summary_base_dir.

    Args:
        data_manager (DataManager): Data manager defined as
            mindinsight.datavisual.data_transform.data_manager.DataManager
        summary_base_dir (str): The summary base directory. It contains summary
            directories generated by training.
        search_condition (dict): The search condition.
    """
    if data_manager is None and summary_base_dir is None:
        raise LineageParamTypeError("One of data_manager or summary_base_dir needs to be specified.")

    if data_manager is None:
        summary_base_dir = validate_and_normalize_path(summary_base_dir, 'summary_base_dir')
    else:
        summary_base_dir = data_manager.summary_base_dir

    search_condition = {} if search_condition is None else search_condition

    try:
        validate_condition(search_condition)
        validate_search_model_condition(SearchModelConditionParameter, search_condition)
    except MindInsightException as error:
        log.error(str(error))
        log.exception(error)
        raise LineageSearchConditionParamError(str(error.message))

    try:
        lineage_objects = LineageOrganizer(data_manager, summary_base_dir).super_lineage_objs
        result = Querier(lineage_objects).filter_summary_lineage(condition=search_condition)
    except LineageSummaryParseException:
        result = {'object': [], 'count': 0}
    except (LineageQuerierParamException, LineageParamTypeError) as error:
        log.error(str(error))
        log.exception(error)
        raise LineageQuerySummaryDataError("Filter summary lineage failed.")

    return result


def get_flattened_lineage(data_manager=None, summary_base_dir=None, search_condition=None):
    """
    Get lineage data in a table from data manager.

    Args:
        data_manager (mindinsight.datavisual.data_manager.DataManager): An object to manage loading.
        summary_base_dir (str): The base directory for train jobs.
        search_condition (dict): The search condition.

    Returns:
        Dict[str, list]: A dict contains keys and values from lineages.

    """
    flatten_dict, user_count = {'train_id': []}, 0
    lineages = filter_summary_lineage(data_manager, summary_base_dir, search_condition).get("object", [])
    for index, lineage in enumerate(lineages):
        flatten_dict['train_id'].append(lineage.get("summary_dir"))
        for key, val in _flatten_lineage(lineage.get('model_lineage', {})):
            if key.startswith(USER_DEFINED_PREFIX) and key not in flatten_dict:
                if user_count > USER_DEFINED_INFO_LIMIT:
                    log.warning("The user_defined_info has reached the limit %s. %r is ignored",
                                USER_DEFINED_INFO_LIMIT, key)
                    continue
                user_count += 1
            if key not in flatten_dict:
                flatten_dict[key] = [None] * index
            flatten_dict[key].append(_parse_value(val))
        for vals in flatten_dict.values():
            if len(vals) == index:
                vals.append(None)
    return flatten_dict


def _flatten_lineage(lineage):
    """Flatten the lineage."""
    for key, val in lineage.items():
        if key == 'metric':
            for k, v in val.items():
                yield f'{METRIC_PREFIX}{k}', v
        elif key == 'user_defined':
            for k, v in val.items():
                yield f'{USER_DEFINED_PREFIX}{k}', v
        else:
            yield key, val


def _parse_value(val):
    """Parse value."""
    if isinstance(val, str) and val.lower() in ['nan', 'inf']:
        return np.nan
    return val


class LineageTable:
    """Wrap lineage data in a table."""
    _LOSS_NAME = "loss"
    _NOT_TUNABLE_NAMES = [_LOSS_NAME, "train_id", "device_num", "model_size",
                          "test_dataset_count", "train_dataset_count"]

    def __init__(self, df: pd.DataFrame):
        self._df = df
        self.train_ids = self._df["train_id"].tolist()
        self._drop_columns_info = []
        self._remove_unsupported_columns()

    def _remove_unsupported_columns(self):
        """Remove unsupported columns."""
        columns_to_drop = []
        for name, data in self._df.iteritems():
            if not is_simple_numpy_number(data.dtype):
                columns_to_drop.append(name)

        if columns_to_drop:
            log.debug("Unsupported columns: %s", columns_to_drop)
            self._df = self._df.drop(columns=columns_to_drop)

        for name in columns_to_drop:
            if not name.startswith(USER_DEFINED_PREFIX):
                continue
            self._drop_columns_info.append({
                "name": name,
                "unselected": True,
                "reason_code": ReasonCode.NOT_ALL_NUMBERS.value
            })

    @property
    def target_names(self):
        """Get names for optimize targets (eg loss, accuracy)."""
        target_names = [name for name in self._df.columns if name.startswith(METRIC_PREFIX)]
        if self._LOSS_NAME in self._df.columns:
            target_names.append(self._LOSS_NAME)
        return target_names

    @property
    def hyper_param_names(self, tunable=True):
        """Get hyper param names."""
        blocked_names = self._get_blocked_names(tunable)

        hyper_param_names = [
            name for name in self._df.columns
            if not name.startswith(METRIC_PREFIX) and name not in blocked_names]

        if self._LOSS_NAME in hyper_param_names:
            hyper_param_names.remove(self._LOSS_NAME)

        return hyper_param_names

    def _get_blocked_names(self, tunable):
        if tunable:
            block_names = self._NOT_TUNABLE_NAMES
        else:
            block_names = []
        return block_names

    @property
    def user_defined_hyper_param_names(self):
        """Get user defined hyper param names."""
        names = [name for name in self._df.columns if name.startswith(USER_DEFINED_PREFIX)]
        return names

    def get_column(self, name):
        """
        Get data for specified column.
        Args:
            name (str): column name.

        Returns:
            np.ndarray, specified column.

        """
        return self._df[name]

    def get_column_values(self, name):
        """
        Get data for specified column.
        Args:
            name (str): column name.

        Returns:
            list, specified column data. If value is np.nan, transform to None.

        """
        return [None if np.isnan(num) else num for num in self._df[name].tolist()]

    @property
    def dataframe_data(self):
        """Get the DataFrame."""
        return self._df

    @property
    def drop_column_info(self):
        """Get dropped columns info."""
        return self._drop_columns_info


def get_lineage_table(data_manager=None, summary_base_dir=None, search_condition=None):
    """Get lineage table from data_manager."""
    lineage_table = get_flattened_lineage(data_manager, summary_base_dir, search_condition)
    lineage_table = LineageTable(pd.DataFrame(lineage_table))

    return lineage_table



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: lazy_read.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='lazy_read.proto',
  package='mindinsight.summary',
  syntax='proto2',
  serialized_options=b'\370\001\001',
  serialized_pb=b'\n\x0flazy_read.proto\x12\x13mindinsight.summary\"t\n\x05\x45vent\x12\x11\n\twall_time\x18\x01 \x02(\x01\x12\x0c\n\x04step\x18\x02 \x01(\x03\x12\x11\n\x07version\x18\x03 \x01(\tH\x00\x12/\n\x07summary\x18\x05 \x01(\x0b\x32\x1c.mindinsight.summary.SummaryH\x00\x42\x06\n\x04what\"\xc8\x01\n\x07Summary\x12\x31\n\x05value\x18\x01 \x03(\x0b\x32\".mindinsight.summary.Summary.Value\x1a\x1e\n\x05Image\x12\x15\n\rencoded_image\x18\x04 \x02(\x0c\x1aj\n\x05Value\x12\x0b\n\x03tag\x18\x01 \x02(\t\x12\x16\n\x0cscalar_value\x18\x03 \x01(\x02H\x00\x12\x33\n\x05image\x18\x04 \x01(\x0b\x32\".mindinsight.summary.Summary.ImageH\x00\x42\x07\n\x05valueB\x03\xf8\x01\x01'
)




_EVENT = _descriptor.Descriptor(
  name='Event',
  full_name='mindinsight.summary.Event',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='wall_time', full_name='mindinsight.summary.Event.wall_time', index=0,
      number=1, type=1, cpp_type=5, label=2,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='step', full_name='mindinsight.summary.Event.step', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='mindinsight.summary.Event.version', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='summary', full_name='mindinsight.summary.Event.summary', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='what', full_name='mindinsight.summary.Event.what',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=40,
  serialized_end=156,
)


_SUMMARY_IMAGE = _descriptor.Descriptor(
  name='Image',
  full_name='mindinsight.summary.Summary.Image',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='encoded_image', full_name='mindinsight.summary.Summary.Image.encoded_image', index=0,
      number=4, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=221,
  serialized_end=251,
)

_SUMMARY_VALUE = _descriptor.Descriptor(
  name='Value',
  full_name='mindinsight.summary.Summary.Value',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='tag', full_name='mindinsight.summary.Summary.Value.tag', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scalar_value', full_name='mindinsight.summary.Summary.Value.scalar_value', index=1,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='image', full_name='mindinsight.summary.Summary.Value.image', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='mindinsight.summary.Summary.Value.value',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=253,
  serialized_end=359,
)

_SUMMARY = _descriptor.Descriptor(
  name='Summary',
  full_name='mindinsight.summary.Summary',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='mindinsight.summary.Summary.value', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_SUMMARY_IMAGE, _SUMMARY_VALUE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=159,
  serialized_end=359,
)

_EVENT.fields_by_name['summary'].message_type = _SUMMARY
_EVENT.oneofs_by_name['what'].fields.append(
  _EVENT.fields_by_name['version'])
_EVENT.fields_by_name['version'].containing_oneof = _EVENT.oneofs_by_name['what']
_EVENT.oneofs_by_name['what'].fields.append(
  _EVENT.fields_by_name['summary'])
_EVENT.fields_by_name['summary'].containing_oneof = _EVENT.oneofs_by_name['what']
_SUMMARY_IMAGE.containing_type = _SUMMARY
_SUMMARY_VALUE.fields_by_name['image'].message_type = _SUMMARY_IMAGE
_SUMMARY_VALUE.containing_type = _SUMMARY
_SUMMARY_VALUE.oneofs_by_name['value'].fields.append(
  _SUMMARY_VALUE.fields_by_name['scalar_value'])
_SUMMARY_VALUE.fields_by_name['scalar_value'].containing_oneof = _SUMMARY_VALUE.oneofs_by_name['value']
_SUMMARY_VALUE.oneofs_by_name['value'].fields.append(
  _SUMMARY_VALUE.fields_by_name['image'])
_SUMMARY_VALUE.fields_by_name['image'].containing_oneof = _SUMMARY_VALUE.oneofs_by_name['value']
_SUMMARY.fields_by_name['value'].message_type = _SUMMARY_VALUE
DESCRIPTOR.message_types_by_name['Event'] = _EVENT
DESCRIPTOR.message_types_by_name['Summary'] = _SUMMARY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Event = _reflection.GeneratedProtocolMessageType('Event', (_message.Message,), {
  'DESCRIPTOR' : _EVENT,
  '__module__' : 'lazy_read_pb2'
  # @@protoc_insertion_point(class_scope:mindinsight.summary.Event)
  })
_sym_db.RegisterMessage(Event)

Summary = _reflection.GeneratedProtocolMessageType('Summary', (_message.Message,), {

  'Image' : _reflection.GeneratedProtocolMessageType('Image', (_message.Message,), {
    'DESCRIPTOR' : _SUMMARY_IMAGE,
    '__module__' : 'lazy_read_pb2'
    # @@protoc_insertion_point(class_scope:mindinsight.summary.Summary.Image)
    })
  ,

  'Value' : _reflection.GeneratedProtocolMessageType('Value', (_message.Message,), {
    'DESCRIPTOR' : _SUMMARY_VALUE,
    '__module__' : 'lazy_read_pb2'
    # @@protoc_insertion_point(class_scope:mindinsight.summary.Summary.Value)
    })
  ,
  'DESCRIPTOR' : _SUMMARY,
  '__module__' : 'lazy_read_pb2'
  # @@protoc_insertion_point(class_scope:mindinsight.summary.Summary)
  })
_sym_db.RegisterMessage(Summary)
_sym_db.RegisterMessage(Summary.Image)
_sym_db.RegisterMessage(Summary.Value)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)


# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""This file is used to define the basic graph."""
import copy
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.datavisual.data_transform.graph.node import Node
from ...domain.graph.base import NodeTypeEnum
from .debugger_graph import DebuggerGraph


class DebuggerMultiGraph(DebuggerGraph):
    """The `DebuggerMultiGraph` object provides interfaces to describe a debugger multigraph."""

    def add_graph(self, graph_dict):
        """
        Add graphs to DebuggerMultiGraph.

        Args:
            graph_dict (dict): The <graph_name, graph_object> dict.
        """
        if len(graph_dict) == 1:
            graph = list(graph_dict.values())[0]
            self._normal_node_map = graph.normal_node_map
            self._node_id_map_name = graph.node_id_map_name
            self._const_node_temp_cache = graph.const_node_temp_cache
            self._parameter_node_temp_cache = graph.parameter_node_temp_cache
            self._leaf_nodes = graph.leaf_nodes
            self._full_name_map_name = graph.full_name_map_name
        else:
            for graph_name, graph in graph_dict.items():
                log.debug("add graph %s into whole graph.", graph_name)

                # add nodes
                normal_nodes = copy.deepcopy(graph.normal_node_map)
                for _, node_obj in normal_nodes.items():
                    self._add_graph_scope(node_obj, graph_name)
                    self._cache_node(node_obj)

                # add graph_node
                node = Node(name=graph_name, node_id=graph_name)
                node.type = NodeTypeEnum.NAME_SCOPE.value
                node.subnode_count = len(graph.list_node_by_scope())
                self._cache_node(node)

            self._leaf_nodes = self._get_leaf_nodes()
            self._full_name_map_name = self._get_leaf_node_full_name_map()

            log.info(
                "Build multi_graph end, all node count: %s, const count: %s, parameter count: %s.",
                self.normal_node_count, len(self._const_node_temp_cache),
                len(self._parameter_node_temp_cache))

    def _add_graph_scope(self, node, graph_name):
        """Add graph scope to the inputs and outputs in node"""

        # add graph scope to node name
        pre_scope = graph_name + "/"
        node.name = pre_scope + node.name
        if node.scope:
            node.scope = pre_scope + node.scope
        else:
            node.scope = graph_name

        # update inputs
        old_inputs = copy.deepcopy(node.inputs)
        for src_name, input_attr in old_inputs.items():
            new_src_name = graph_name + "/" + src_name
            node.add_inputs(new_src_name, input_attr)
            node.delete_inputs(src_name)

        # update outputs
        old_outputs = copy.deepcopy(node.outputs)
        for dst_name, output_attr in old_outputs.items():
            new_dst_name = graph_name + "/" + dst_name
            node.add_outputs(new_dst_name, output_attr)
            node.delete_outputs(dst_name)

        # update proxy_inputs
        old_proxy_inputs = copy.deepcopy(node.proxy_inputs)
        for src_name, input_attr in old_proxy_inputs.items():
            new_src_name = graph_name + "/" + src_name
            node.add_proxy_inputs(new_src_name, input_attr)
            node.delete_proxy_inputs(src_name)

        # update proxy_outputs
        old_proxy_outputs = copy.deepcopy(node.proxy_outputs)
        for dst_name, output_attr in old_proxy_outputs.items():
            new_dst_name = graph_name + "/" + dst_name
            node.add_proxy_outputs(new_dst_name, output_attr)
            node.delete_proxy_outputs(dst_name)



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Command module."""

import os
import sys
import stat
import argparse
from importlib import import_module

import mindinsight
from mindinsight.utils.log import setup_logger
from mindinsight.utils.exceptions import MindInsightException


class BaseCommand:
    """Base command class."""

    name = ''
    description = ''

    # logger for console output instead of built-in print
    console = None

    # logger for log file recording in case audit is required
    logfile = None

    def add_arguments(self, parser):
        """
        Add arguments to parser.

        Args:
            parser (ArgumentParser): specify parser to which arguments are added.
        """

    def update_settings(self, args):
        """
        Update settings.

        Args:
            args (Namespace): parsed arguments to hold customized parameters.
        """

    def run(self, args):
        """
        Implementation of command logic.

        Args:
            args (Namespace): parsed arguments to hold customized parameters.
        """
        raise NotImplementedError('subclasses of BaseCommand must provide a run() method')

    def invoke(self, args):
        """
        Invocation of command.

        Args:
            args (Namespace): parsed arguments to hold customized parameters.
        """
        error = None
        try:
            self.update_settings(args)
        except MindInsightException as e:
            error = e

        self.console = setup_logger('mindinsight', 'console', console=True, logfile=False, formatter='%(message)s')
        if error is not None:
            self.console.error(error.message)
            sys.exit(1)

        self.logfile = setup_logger('scripts', self.name, console=False, logfile=True)
        self.run(args)


def main():
    """Entry point for mindinsight CLI."""

    console = setup_logger('mindinsight', 'console', console=True, logfile=False, formatter='%(message)s')
    if (sys.version_info.major, sys.version_info.minor) < (3, 7):
        console.error('Python version should be at least 3.7')
        sys.exit(1)

    # set umask to 0o077
    os.umask(stat.S_IRWXG | stat.S_IRWXO)

    parser = argparse.ArgumentParser(
        prog='mindinsight',
        description='MindInsight CLI entry point (version: {})'.format(mindinsight.__version__),
        allow_abbrev=False)

    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s ({})'.format(mindinsight.__version__))

    subparsers = parser.add_subparsers(
        dest='cli',
        title='subcommands',
        description='the following subcommands are supported',
    )

    commands = {}
    scripts_path = os.path.realpath(os.path.join(__file__, os.pardir, os.pardir, 'scripts'))
    files = os.listdir(scripts_path)
    files.sort()
    for file in files:
        if file.startswith('_') or not file.endswith('.py'):
            continue

        module = import_module('mindinsight.scripts.{}'.format(file[:-len('.py')]))
        command_cls = getattr(module, 'Command', None)
        if command_cls is None or not issubclass(command_cls, BaseCommand):
            continue

        command = command_cls()
        command_parser = subparsers.add_parser(command.name, help=command.description, allow_abbrev=False)
        command.add_arguments(command_parser)
        commands[command.name] = command

    argv = sys.argv[1:]
    if not argv or argv[0] == 'help':
        argv = ['-h']

    args = parser.parse_args(argv)
    cli = args.__dict__.pop('cli')
    command = commands[cli]
    command.invoke(args)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""This file is used to define the MindSpore graph."""
from collections import defaultdict

from mindinsight.datavisual.common.log import logger
from mindinsight.datavisual.data_transform.graph.msgraph import MSGraph
from mindinsight.domain.graph.base import NodeTypeEnum


class OptimizedGraph(MSGraph):
    """The object describes the MindSpore graph, and it is defined in the anf_ir proto file."""

    MIN_GROUP_NODE_COUNT = 10

    def __init__(self):
        super().__init__()
        self._load_node_temp_cache = {}

    def _inherit_input_output_from_subnode(self, parent_node, subnode_list, filtered_type=None):
        """
        Adds the input and output of all direct child nodes to the current node.

        Args:
            parent_node (Node): The nodes that inherit the input and output of the child nodes.
            subnode_list (list[Node]): A list of child nodes that are inherited from the input and output.
            filtered_type (set(str)): Filter some input and output that do not require inheritance
                                      based on the node type. Default is filter const node.

        Note:
            - Only the inputs and outputs of the external scope are inherited.
            - Before add_const_node method, if the input is a const,
              the scope of the const node is not startswith the name of parent node.
              So in this scenario, we need to filter the const nodes.
        """
        filtered_type = {NodeTypeEnum.CONST.value} if filtered_type is None else filtered_type
        for method in ['inputs', 'outputs', 'proxy_inputs', 'proxy_outputs']:
            for node in subnode_list:
                for item_name, item_attr in getattr(node, method).items():
                    target_node = self._get_normal_node(node_name=item_name)
                    if target_node is None:
                        logger.warning("inherit %s from subnode, target node (%s) is None", method, item_name)
                        continue

                    if item_name.startswith(f'{parent_node.name}/'):
                        continue

                    if target_node.type in filtered_type:
                        continue

                    getattr(parent_node, f'add_{method}')(item_name, item_attr)

    def _cache_node(self, node):
        """Store the node in the cache."""
        # Notice:
        # The additional caching is used to handle the Const, Parameter and LOAD nodes separately later.
        super()._cache_node(node)
        if node.type == NodeTypeEnum.LOAD.value:
            self._load_node_temp_cache.update({node.name: node})

    def _delete_nodes_of_cache(self, node_names):
        """Delete node from cache."""
        logger.debug("These nodes will be removed from the cache, node names: %s.", node_names)
        for name in node_names:

            if self._parameter_node_temp_cache.get(name):
                self._parameter_node_temp_cache.pop(name)
            if self._const_node_temp_cache.get(name):
                self._const_node_temp_cache.pop(name)
            if self._load_node_temp_cache.get(name):
                self._load_node_temp_cache.pop(name)

            node = self._get_normal_node(node_name=name)
            self._normal_node_map.pop(name)
            self._node_id_map_name.pop(node.node_id)

    def _parse_data(self, proto_data):
        """
        The proto data is parsed and all nodes are stored in the specified structure.

        Args:
            proto_data (anf_ir_pb2.GraphProto): Refer to anf_ir_pb2.GraphProto object.
        """
        logger.info("Start to parse graph proto data.")

        self._parse_op_nodes(proto_data.node)
        self._parse_parameters(proto_data.parameters)
        self._parse_consts(proto_data.const_vals)

        self._update_input_after_create_node()
        self._update_output_after_create_node()

        self._delete_non_computational_ops()
        self._clean_no_input_output_node()

        self._extract_node_by_single_node_in_scope()
        logger.info("Parse proto data end, normal node count(only contain op node, "
                    "parameter, const): %s.", self.normal_node_count)

    def _parse_op_nodes(self, node_protos):
        """
        Parse `anf_ir_pb2.NodeProto` object, and create a normal node.

        Args:
            node_protos (list[anf_ir_pb2.NodeProto]): Refer to anf_ir_pb2.NodeProto.
        """
        logger.debug("Start to parse op nodes from proto.")
        for topological_index, node_proto in enumerate(node_protos):
            if not node_proto.name:
                logger.warning("Finding a node with an empty name will not save it.")
                continue

            if node_proto.full_name.startswith("Gradients") or "optimizer" in node_proto.full_name \
                    or "opt" in node_proto.instance_name:
                continue

            self._parse_op_node(topological_index, node_proto)

    def _update_input_after_create_node(self):
        """Update the input of node after create node."""
        for node in self._normal_node_map.values():
            for src_node_id, input_attr in dict(node.inputs).items():
                node.delete_inputs(src_node_id)
                if not self._is_node_exist(node_id=src_node_id):
                    continue

                src_node = self._get_normal_node(node_id=src_node_id)
                input_attr['shape'] = src_node.output_shape
                input_attr['data_type'] = src_node.output_data_type
                node.add_inputs(src_name=src_node.name, input_attr=input_attr)

        nodes = self._list_nodes_without_parameter_const()
        for node in nodes:
            for src_node_name, _ in dict(node.inputs).items():
                if not self._is_node_exist(node_name=src_node_name):
                    logger.warning("Source node (%s) is None.", src_node_name)
                    continue

                src_node = self._get_normal_node(node_name=src_node_name)
                if src_node.type in (NodeTypeEnum.LOAD.value,
                                     NodeTypeEnum.TUPLE_GET_ITEM.value,
                                     NodeTypeEnum.MAKETUPLE.value,
                                     NodeTypeEnum.UPDATE_STATE.value):
                    node.delete_inputs(src_node_name)
                    for source_node_name, source_attr in dict(src_node.inputs).items():
                        source_node = self._get_normal_node(node_name=source_node_name)
                        if source_node is None:
                            logger.warning("Source node (%s) is None.", source_node_name)
                            continue
                        source_attr['shape'] = source_node.output_shape
                        source_attr['data_type'] = source_node.output_data_type
                        node.add_inputs(src_name=source_node.name, input_attr=source_attr)

    def _update_output_after_create_node(self):
        """Update the output of node after create node."""
        super()._update_output_after_create_node()

        nodes = self._list_nodes_without_parameter_const()
        for node in nodes:
            for src_node_name, _ in dict(node.outputs).items():
                if not self._is_node_exist(node_name=src_node_name):
                    logger.warning("Source node (%s}) is None.", src_node_name)
                    continue

                src_node = self._get_normal_node(node_name=src_node_name)
                if src_node.type in (NodeTypeEnum.LOAD.value,
                                     NodeTypeEnum.TUPLE_GET_ITEM.value,
                                     NodeTypeEnum.MAKETUPLE.value,
                                     NodeTypeEnum.UPDATE_STATE.value):
                    node.delete_outputs(src_node_name)
                    for source_node_name, source_attr in dict(src_node.outputs).items():
                        source_node = self._get_normal_node(node_name=source_node_name)
                        if source_node is None:
                            logger.warning("Source node (%s) is None.", source_node_name)
                            continue
                        source_attr['shape'] = source_node.output_shape
                        source_attr['data_type'] = source_node.output_data_type
                        node.add_outputs(src_name=source_node.name, output_attr=source_attr)

    def _delete_non_computational_ops(self):
        """Deleted non-computational operators."""
        delete_names = []
        for node in self._normal_node_map.values():
            if node.type in (NodeTypeEnum.LOAD.value,
                             NodeTypeEnum.TUPLE_GET_ITEM.value,
                             NodeTypeEnum.MAKETUPLE.value,
                             NodeTypeEnum.UPDATE_STATE.value):
                delete_names.append(node.name)

        self._delete_nodes_of_cache(delete_names)

    def _list_nodes_without_parameter_const(self):
        """List nodes without parameter and const node."""
        nodes = self._normal_node_map.values()
        not_expect_type = (NodeTypeEnum.CONST.value, NodeTypeEnum.PARAMETER.value)
        nodes = filter(lambda node: node.type not in not_expect_type, nodes)
        nodes = sorted(nodes, key=lambda node: node.topological_index)
        return nodes

    def _extract_node_by_single_node_in_scope(self):
        """Extract node from the scope which has only one node."""
        nodes = self._list_nodes_without_parameter_const()
        scope_map_types = defaultdict(set)
        scope_map_node_cnt = defaultdict(int)
        for node in nodes:
            if not node.scope or '/' not in node.scope:
                continue
            scope_map_types[node.scope].add(node.type)
            scope_map_node_cnt[node.scope] += 1

        filter_scopes = set()
        for scope, types in scope_map_types.items():
            if len(types) == 1 and scope_map_node_cnt[scope] > 1 and types.pop() in scope:
                filter_scopes.add(scope)

        for filter_scope in list(filter_scopes):
            for scope in scope_map_types:
                if scope.startswith(f'{filter_scope}/'):
                    filter_scopes.remove(filter_scope)
                    break

        if not filter_scopes:
            return

        for node in nodes:
            if node.scope in filter_scopes and '/' in node.scope:
                name = node.name.rsplit('/', 1)[1]
                new_scope = node.scope.rsplit('/', 1)[0]
                new_name = f'{new_scope}/{name}'
                self._update_node_name_of_cache(node, new_name)

        return

    def _clean_no_input_output_node(self):
        """Clean nodes which has no input and output."""
        nodes = self._list_nodes_without_parameter_const()
        deleted_names = []
        for node in nodes:
            if not node.inputs and not node.outputs:
                deleted_names.append(node.name)
        self._delete_nodes_of_cache(deleted_names)



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
MindWizard.

MindWizard is a tool for quickly generating classic network scripts.
It collects user preference on the combination of network parameters,
such models, hyperparameters and datasets, then automatically generates target network scripts.
The generated scripts can be used for training and evaluation in the Ascend or GPU environment.
"""



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The Compose Proposals."""
from collections import OrderedDict

from mindinsight.profiler.common.log import logger
from mindinsight.profiler.common.util import get_options
from mindinsight.profiler.proposer.proposer_factory import ProposerFactory


class ComposeProposal:
    """Get the proposals from multiple different proposers."""

    def __init__(self, profiling_path, device_id, type_list=None):
        self.profiling_path = profiling_path
        self.device_id = device_id
        self.compose_proposer_type_list = type_list
        # Postfix of category label, used for UI to identify the label as category label.
        self.type_label_postfix = "-proposer_type_label"

    def get_proposal(self, options=None):
        """
        Get compose proposals.

        Args:
            options (dict): options for composed proposal.

                - compose_proposal_result: execution results of the already running proposers.
                - step_trace: include optional parameters for step trace，The dictionary key is iter_interval
                  used to get the analyser options for iteration interval time.

        Returns:
            dict, the proposals from multiple different proposers.

        Examples:
            >>> type_list = ['common', 'step_trace']
            >>> condition = {"filter_condition": {'mode': "proc", "proc_name": "iteration_interval"}}
            >>> options = {'step_trace': {"iter_interval": condition}}
            >>> cp = ComposeProposal(self.profiling_dir, self.device_id, type_list)
            >>> result_proposal = cp.get_proposal(options=options)
        """
        logger.info("The ComposeProposal is running")
        options = get_options(options)
        logger.debug("The 'options'　is %s", str(options))
        # The flag whether to write category label.
        type_label_flag = options.get("type_label_flag", True)
        compose_proposal_result = OrderedDict()
        logger.debug("The 'compose_proposer_type_list'　is %s", str(self.compose_proposer_type_list))
        for proposer_type in self.compose_proposer_type_list:
            proposer = ProposerFactory.instance().get_proposer(proposer_type, self.profiling_path, self.device_id)

            if proposer is None:
                continue

            # Write the result of proposals to option for other proposer to get.
            options["compose_proposal_result"] = compose_proposal_result

            result = proposer.analyze(options)
            # Insert category label.
            if result and type_label_flag:
                proposer_type_label = proposer_type + "-type_label"
                # Get the name of the category label, the default is the same as the proposer type.
                type_label_name = options.get(proposer_type_label, proposer_type)
                # Add postfix to category label name
                type_proposal_label = type_label_name + self.type_label_postfix
                compose_proposal_result[type_proposal_label] = None
                # Merge results to the proposals dictionary.
                compose_proposal_result.update(result)
            elif result and not type_label_flag:
                # Merge results to the proposals dictionary.
                compose_proposal_result.update(result)
        compose_proposal_result = self.show_result_condition(compose_proposal_result)
        logger.debug("The 'compose_proposal_result' is %s", str(compose_proposal_result))
        return compose_proposal_result

    def show_result_condition(self, proposal_result):
        """Hide some suggestions based on conditions."""

        # If the average step interval is greater than StepTraceProposer.__step_trace_iter_interval_threshold or the
        # data processing performance optimization is empty, it is recommended that only the Profiling and Optimization
        # Guide be displayed.
        hidden_key = list()
        if 'step_trace-iter_interval' not in proposal_result.keys() \
                or 'minddata_pipeline-general' not in proposal_result.keys():
            for key in proposal_result.keys():
                if key not in ['common-proposer_type_label', 'common-profiler_tutorial', 'msadvisor-proposer']:
                    hidden_key.append(key)
        else:
            # Hide optimization suggestions for data processing if there is a problem with data transmission.
            empty_rate = proposal_result.get('minddata_device_queue_rate', {}).get('empty_rate')
            empty_warning_threshold = proposal_result.get('minddata_device_queue_rate', {})\
                .get('empty_warning_threshold')
            if empty_rate and empty_rate <= empty_warning_threshold:
                for key in proposal_result.keys():
                    if key.startswith('minddata_pipeline-') and key not in hidden_key:
                        hidden_key.append(key)

        for key in hidden_key:
            del proposal_result[key]
        return proposal_result



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Management of all conditions.

This module is used to register all conditions, as well as their parameters.
This module also provide the available conditions to condition_collections api.
"""
from enum import Enum
from mindinsight.debugger.common.log import LOGGER as logger


class ParamNameEnum(Enum):
    """Param names."""
    ABS_MEAN_GT = "abs_mean_gt"
    ABS_MEAN_LT = "abs_mean_lt"
    ABS_MEAN_UPDATE_RATIO_GT = "abs_mean_update_ratio_gt"
    ABS_MEAN_UPDATE_RATIO_LT = "abs_mean_update_ratio_lt"
    ATOL = "atol"
    EQUAL_NAN = "equal_nan"
    EPSILON = "epsilon"
    MAX_GT = "max_gt"
    MAX_LT = "max_lt"
    MIN_GT = "min_gt"
    MIN_LT = "min_lt"
    MEAN_GT = "mean_gt"
    MEAN_LT = "mean_lt"
    MAX_MIN_GT = "max_min_gt"
    MAX_MIN_LT = "max_min_lt"
    PARAM = "param"
    RANGE_START_INCLUSIVE = "range_start_inclusive"
    RANGE_END_INCLUSIVE = "range_end_inclusive"
    RANGE_PERCENTAGE_GT = "range_percentage_gt"
    RANGE_PERCENTAGE_LT = "range_percentage_lt"
    RTOL = "rtol"
    ZERO_PERCENTAGE_GE = "zero_percentage_ge"


class ConditionIdEnum(Enum):
    """Condition ids."""
    WEIGHT_INITIALIZATION = "weight_initialization"
    WEIGHT_OVERFLOW = "weight_overflow"
    WEIGHT_TOO_LARGE = "weight_too_large"
    WEIGHT_TOO_SMALL = "weight_too_small"
    GRADIENT_VANISHING = "gradient_vanishing"
    GRADIENT_TOO_LARGE = "gradient_too_large"
    GRADIENT_EXPLODING = "gradient_exploding"
    TENSOR_OVERFLOW = "tensor_overflow"
    OPERATOR_OVERFLOW = "operator_overflow"
    TENSOR_TOO_LARGE = "tensor_too_large"
    TENSOR_TOO_SMALL = "tensor_too_small"
    TENSOR_ALL_ZERO = "tensor_all_zero"
    WEIGHT_NOT_CHANGED = "weight_not_changed"
    WEIGHT_CHANGE_TOO_LARGE = "weight_change_too_large"
    WEIGHT_CHANGE_TOO_SMALL = "weight_change_too_small"
    ACTIVATION_RANGE = "activation_range"
    TENSOR_RANGE = "tensor_range"


class OptimizePhaseEnum(Enum):
    """Optimize phases."""
    TENSOR_CHECK = 400
    OPERATOR_CHECK = 100
    LOSS_CHECK = 300
    INPUT_DATA_CHECK = 200


class ValueTypeEnum(Enum):
    """Value types."""
    FLOAT64 = 1
    INT64 = 2
    BOOL = 3


class PlatformEnum(Enum):
    """Platform types."""
    GPU = "GPU"
    ASCEND = "Ascend"


class TargetTypeEnum(Enum):
    """Target types."""
    TENSOR = 'tensor'
    ACTIVATION = 'activation'
    GRADIENT = 'gradient'
    PARAMETER = 'parameter'
    WEIGHT = 'weight'


class ParamTypeEnum(Enum):
    """Param types."""
    CHECK_PARAM = "CHECK_PARAM"
    SUPPORT_PARAM = "SUPPORT_PARAM"


class ActivationFuncEnum(Enum):
    """Activation functions."""
    TANH = 'tanh'
    SIGMOID = 'sigmoid'
    RELU = 'relu'
    RELUV2 = 'reluv2'


class ConditionContext:
    """
    The class for condition context.

    Args:
        backend (str): parameter name.
        step (int): the type of value.
        debugger_capability (tuple): whether the param support no assignment.
    """
    def __init__(self, backend, step=0, debugger_capability=(1, 1)):
        self._backend = backend
        self._step = step
        self._debugger_capability = debugger_capability

    @property
    def backend(self):
        """Get backend."""
        return self._backend

    @property
    def step(self):
        """Get _step."""
        return self._step

    @property
    def debugger_capability(self):
        """Get debugger_capability."""
        return self._debugger_capability


class ConditionParameter:
    """
    The class for parameters of conditions.

    Args:
        name (ParamNameEnum): parameter name.
        value_type (ValueTypeEnum): the type of value.
        valid_test_func (func): the function used to test whether the param is valid.
        support_disable (bool): whether the param support no assignment.
        default_value (float): default value.
        visible_on_ui (bool): whether the param visible on ui.
        param_type (ParamTypeEnum): parameters type.
        required_params (list): the list of required parameters.
    """

    def __init__(self, name, value_type: ValueTypeEnum, valid_test_func=None, support_disable=True, default_value=None,
                 visible_on_ui=True, param_type=ParamTypeEnum.CHECK_PARAM, required_params=None):
        self._name = name.value
        self._type = value_type
        self._valid_test_func = valid_test_func
        self._support_disable = support_disable
        self._default_value = default_value
        self._visible_on_ui = visible_on_ui
        self._param_type = param_type.value
        self._required_params = required_params

    @property
    def name(self):
        """Get name of parameter."""
        return self._name

    @property
    def type(self):
        """Get type of parameter."""
        return self._type

    @property
    def support_disable(self):
        """Get support_disable of parameter."""
        return self._support_disable

    @property
    def default_value(self):
        """Get default_value of parameter."""
        return self._default_value

    @property
    def visible_on_ui(self):
        """Get visible_on_ui of parameter."""
        return self._visible_on_ui

    @property
    def param_type(self):
        """Get param_type of parameter."""
        return self._param_type

    @property
    def required_params(self):
        """Get required_param of parameter."""
        return self._required_params

    def is_valid(self, value):
        """Check is the parameter valid."""
        if self._valid_test_func is None:
            return True
        return self._valid_test_func(value)


class Condition:
    """
    The class for parameters of conditions.

    Args:
        condition_id (ConditionIdEnum): condition id.
        abbr (str): the abbreviation of condition id.
        optimize_phase (OptimizePhaseEnum): optimize phase.
        parameters (List[ConditionParameter]): parameters.
        supported_target_type (TargetTypeEnum): the supported target type.
        supported_platforms (tuple[PlatformEnum, PlatformEnum]): the supported platforms.
        minimum_debugger_capability (tuple): the minimum debugger capability required.
        availability_test_func (func): the function used to test whether the condition is available.
    """
    def __init__(self, condition_id, abbr, optimize_phase, parameters, supported_target_type, supported_platforms,
                 minimum_debugger_capability, availability_test_func=None):
        self.id = condition_id.value
        self._abbr = abbr
        self.optimize_phase = optimize_phase
        self._parameters = {
            parameter.name: parameter for parameter in parameters
        }
        self.ordered_parameter_names = [parameter.name for parameter in parameters]
        self._supported_target_type = supported_target_type
        self.supported_platforms = supported_platforms
        self.minimum_debugger_capability = minimum_debugger_capability
        self.availability_test_func = availability_test_func

    def get_parameter_definition(self, name):
        """Return parameter definition by the name"""
        return self._parameters[name]

    def is_available(self, condition_context):
        """Check is the condition available."""
        backend = condition_context.backend
        debugger_capability = condition_context.debugger_capability
        if debugger_capability < self.minimum_debugger_capability:
            logger.debug("The debugger capability is lower than the minimum debugger capability.")
            return False
        if backend not in [platform.value for platform in self.supported_platforms]:
            logger.debug("The condition %s is not supported on the platform.", self.id)
            return False
        if self.availability_test_func is None:
            return True
        return self.availability_test_func(condition_context)

    @property
    def abbr(self):
        """The abbreviation of condition"""
        return self._abbr

    @property
    def names(self):
        """The name of condition"""
        return self._parameters.keys()

    @property
    def parameters(self):
        """The parameters of condition"""
        return self._parameters.values()

    @property
    def supported_target_type(self):
        """The supported target type of condition"""
        return self._supported_target_type


def check_initialization_available(condition_context):
    """Check if initialization is available at this step"""
    if condition_context.step == 0:
        return True
    return False


def check_percentage_param_range(value):
    if 0 <= value <= 100:
        return True
    return False


def check_normal_param_range(value):
    if float("-inf") < value < float("inf"):
        return True
    return False


def check_abs_param_range(value):
    if 0 <= value < float("inf"):
        return True
    return False


def check_positive_param_range(value):
    if 0 < value < float("inf"):
        return True
    return False



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Graph API module."""

import os

from mindinsight.debugger.dump.convert import DumpRootDirConverter, FileMapping
from mindinsight.debugger.dump.parser import Parser
from mindinsight.debugger.stream_cache.data_loader import DataLoader
from mindinsight.domain.graph.query import StackQuery


class DumpParser(Parser):
    """
    Dump Parser.

    Args:
        dump_dir (str): Dump directory path.
    """

    def __init__(self, dump_dir):
        loader = DataLoader(dump_dir)
        super(DumpParser, self).__init__(loader)

    def get_tensor_files(self, qs, use_regex=False, rank_ids=None, iterations=None):
        """
        Get tensor files.

        Args:
            qs (str): Query string.
            use_regex (bool): Indicates if query is regex.
            rank_ids (list): Selected rank IDs.
            iterations (list): Selected iterations.

        Returns:
            dict, paths of tensor files. The format is like: {[op_full_name]: OpPathManager}.
                Call OpPathManager.to_dict(), the result is format like:
                    {'rank_[rank_id]':
                        {[iteration_id]:
                            {
                                'input': list[file_path],
                                'output': list[file_path]
                            }
                        }
                    }
        """
        operators = []
        if rank_ids is None:
            operators = self._operators
        else:
            for operator in self._operators:
                if operator.rank_id in rank_ids:
                    operators.append(operator)

        query = StackQuery(operators)
        operators = query.filter(qs, use_regex=use_regex).all()
        file_paths = {}
        file_mapping = FileMapping(self._loader)
        for operator in operators:
            op_pattern = os.path.basename(operator.full_name)
            res = file_mapping.find_tensor_file(op_pattern, rank_ids=rank_ids, iterations=iterations)
            file_paths[operator.full_name] = res
        return file_paths

    def convert_all_data_to_host(self):
        """
        Convert all data to host format.

        Returns:
            list, failed tensors.
        """
        conversion = DumpRootDirConverter(self._loader)
        failed_lines = conversion.convert()
        return failed_lines



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Command module."""
import argparse
import os
import sys

import mindinsight
from mindinsight.optimizer.tuner import Tuner


class ConfigAction(argparse.Action):
    """Summary base dir action class definition."""

    def __call__(self, parser_in, namespace, values, option_string=None):
        """
        Inherited __call__ method from argparse.Action.

        Args:
            parser_in (ArgumentParser): Passed-in argument parser.
            namespace (Namespace): Namespace object to hold arguments.
            values (object): Argument values with type depending on argument definition.
            option_string (str): Option string for specific argument name.
        """
        config_path = os.path.realpath(values)
        if not os.path.exists(config_path):
            parser_in.error(f'{option_string} {config_path} not exists.')

        setattr(namespace, self.dest, config_path)


class IterAction(argparse.Action):
    """Summary base dir action class definition."""

    def __call__(self, parser_in, namespace, values, option_string=None):
        """
        Inherited __call__ method from argparse.Action.

        Args:
            parser_in (ArgumentParser): Passed-in argument parser.
            namespace (Namespace): Namespace object to hold arguments.
            values (object): Argument values with type depending on argument definition.
            option_string (str): Option string for specific argument name.
        """
        iter_times = values
        if iter_times <= 0:
            parser_in.error(f'{option_string} {iter_times} should be a positive integer.')

        setattr(namespace, self.dest, iter_times)


parser = argparse.ArgumentParser(
        prog='mindoptimizer',
        description='MindOptimizer CLI entry point (version: {})'.format(mindinsight.__version__))

parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s ({})'.format(mindinsight.__version__))

parser.add_argument(
        '--config',
        type=str,
        action=ConfigAction,
        required=True,
        default=os.path.join(os.getcwd(), 'output'),
        help="Specify path for config file."
)

parser.add_argument(
        '--iter',
        type=int,
        action=IterAction,
        default=1,
        help="Optional, specify run times for the command in config file."
)


def cli_entry():
    """Cli entry."""
    argv = sys.argv[1:]
    if not argv:
        argv = ['-h']
        args = parser.parse_args(argv)
    else:
        args = parser.parse_args()

    tuner = Tuner(args.config)
    tuner.optimize(max_expr_times=args.iter)



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Create project command module."""
import os
import re
import sys
import textwrap
from pathlib import Path

import click

from mindinsight.utils.command import BaseCommand
from mindinsight.wizard.base.utility import find_network_maker_names, load_network_maker, process_prompt_choice
from mindinsight.wizard.common.exceptions import CommandError
from mindinsight.wizard.conf.constants import SUPPORT_MINDSPORE_VERSION, QUESTION_START


class CreateProject(BaseCommand):
    """Create project class."""
    name = 'createproject'
    description = 'create project'

    def __init__(self):
        self._network_types = find_network_maker_names()

    def add_arguments(self, parser):
        """
        Add arguments to parser.

        Args:
            parser (ArgumentParser): Specify parser to which arguments are added.
        """
        parser.add_argument(
            'name',
            type=str,
            help='Specify the new project name.')

    def _make_project_dir(self, project_name):
        self._check_project_dir(project_name)
        permissions = os.R_OK | os.W_OK | os.X_OK
        mode = permissions << 6
        project_dir = os.path.join(os.getcwd(), project_name)
        os.makedirs(project_dir, mode=mode, exist_ok=True)
        return project_dir

    @staticmethod
    def _check_project_dir(project_name):
        """Check project directory whether empty or exist."""
        if not re.search('^[A-Za-z0-9][A-Za-z0-9._-]*$', project_name):
            raise CommandError("'%s' is not a valid project name. Please input a valid name matching "
                               "regex ^[A-Za-z0-9][A-Za-z0-9._-]*$" % project_name)
        project_dir = os.path.join(os.getcwd(), project_name)
        if os.path.exists(project_dir):
            output_path = Path(project_dir)
            if output_path.is_dir():
                if os.path.os.listdir(project_dir):
                    raise CommandError('%s already exists, %s is not empty directory, please try another name.'
                                       % (project_name, project_dir))
            else:
                CommandError('There is a file in the current directory has the same name as the project %s, '
                             'please try another name.' % project_name)
        return True

    def ask_network(self):
        """Ask user question for selecting a network to create."""
        network_type_choices = self._network_types[:]
        network_type_choices.sort(reverse=False)
        prompt_msg = '{}:\n{}\n'.format(
            '%sPlease select a network' % QUESTION_START,
            '\n'.join(f'{idx: >4}: {choice}' for idx, choice in enumerate(network_type_choices, start=1))
        )
        prompt_type = click.IntRange(min=1, max=len(network_type_choices))
        choice = 0
        while not choice:
            choice = click.prompt(prompt_msg, default=0, type=prompt_type,
                                  hide_input=False, show_choices=False,
                                  confirmation_prompt=False, show_default=False,
                                  value_proc=lambda x: process_prompt_choice(x, prompt_type))
            if not choice:
                click.secho(textwrap.dedent("Network is required."), fg='red')

        click.secho(textwrap.dedent("Your choice is %s." % network_type_choices[choice - 1]), fg='yellow')
        return network_type_choices[choice - 1]

    @staticmethod
    def echo_notice():
        """Echo notice for depending environment."""
        click.secho(textwrap.dedent(
            "[NOTICE] The final generated scripts should be run under environment "
            "where mindspore==%s and related device drivers are installed. " % SUPPORT_MINDSPORE_VERSION), fg='yellow')

    def run(self, args):
        """Override run method to start."""
        project_name = args.get('name')
        try:
            self._check_project_dir(project_name)
        except CommandError as error:
            click.secho(error.message, fg='red')
            sys.exit(1)
        try:
            self.echo_notice()
            network_maker_name = self.ask_network()
            network_maker = load_network_maker(network_maker_name)
            network_maker.configure()
        except click.exceptions.Abort:
            sys.exit(1)

        project_dir = self._make_project_dir(project_name)
        source_files = network_maker.generate(**args)
        for source_file in source_files:
            source_file.write(project_dir)

        click.secho(f"{project_name} is generated in {project_dir}")



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""DebuggerEngine."""

from mindinsight.debugger.debugger_services.debugger_offline_server import DebuggerOfflineManager


class DebuggerEngine:
    """Debugger Engine object."""

    def __init__(self, data_loader, mem_limit):
        """Initialization."""
        self._data_loader = data_loader
        DebuggerOfflineManager.check_toolkit(self._data_loader)
        self._dbg_services_module = DebuggerOfflineManager.get_dbg_service_module()
        self._dbg_service = DebuggerOfflineManager.get_dbg_service(self.dbg_services_module,
                                                                   data_loader,
                                                                   mem_limit)

    @property
    def data_loader(self):
        """DataLoader object."""
        return self._data_loader

    @property
    def dbg_services_module(self):
        """Get dbg_services module in MindSpore."""
        return self._dbg_services_module

    @property
    def dbg_service(self):
        """Get initialized DbgServices object."""
        return self._dbg_service



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The minddata proposer."""
import math
import os

from collections import OrderedDict

from mindinsight.profiler.analyser.analyser_factory import AnalyserFactory
from mindinsight.profiler.analyser.minddata_analyser import MinddataAnalyser
from mindinsight.profiler.proposer.allproposers.base_proposer import Proposer
from mindinsight.profiler.common.log import logger as log
from mindinsight.profiler.common.exceptions.exceptions import ProfilerRawFileException, ProfilerFileNotFoundException

DATA_PROCESS_WARNING = 'data_process'
DATA_TRANSMISSION_WARNING = 'data_transmission'
DEVICE_QUEUE_WARNING = 'device_queue_warning'


class MinddataProposer(Proposer):
    """The Minddata proposer."""

    def __init__(self, profiling_dir, rank_id):
        super().__init__(profiling_dir, rank_id)
        self.__proposer_type = "minddata"
        self.__proposal_dict = OrderedDict()
        self.empty_warning_threshold = int(MinddataAnalyser.DEVICE_QUEUE_EMPTY_WARNING_THRESHOLD * 100)

    def analyze(self, options=None):
        """
        Get the proposal from proposer.

        Args:
            options (dict): The options for proposer analysis.

        Returns:
            dict, the proposal from proposer instance，the dictionary key is a language internationalization
            label, and the value is used to format the value in the language internationalization string.

        Examples:
            >>> proposer_type = 'minddata'
            >>> proposer = ProposerFactory.instance().get_proposer(proposer_type, self.profiling_dir, self.rank_id)
            >>> result = proposer.analyze(options)
        """
        self.minddata_outer_bounds_analyze()
        self.minddata_cpu_utilization_proposal()
        return self.__proposal_dict

    def minddata_outer_bounds_analyze(self):
        """Get the proposals of minddata outer bounds."""
        minddata_dict = OrderedDict()
        minddata_analyser = AnalyserFactory.instance().get_analyser(
            'minddata', self.profiling_path, self.rank_id)
        get_next_queue_info, _ = minddata_analyser.analyse_get_next_info(info_type="queue")
        device_queue_info, _ = minddata_analyser.analyse_device_queue_info(info_type="queue")

        result = MinddataAnalyser.analyse_queue_summary(get_next_queue_info, device_queue_info)
        if "get_next_queue_info" in result:
            get_next_queue_info_summary = result.get("get_next_queue_info").get("summary", {})
            empty_batch = get_next_queue_info_summary.get("empty_batch_count")
            total_batch = get_next_queue_info_summary.get("total_batch")

            minddata_dict["minddata_get_next_queue"] = [empty_batch, total_batch]
            self.__proposal_dict.update(minddata_dict)
        if "device_queue_info" in result:
            get_next_queue_info_summary = result.get("device_queue_info").get("summary", {})
            full_batch = get_next_queue_info_summary.get("full_batch_count", 0)
            empty_batch = get_next_queue_info_summary.get("empty_batch_count", 0)
            total_batch = get_next_queue_info_summary.get("total_batch", 0)
            device_queue_empty_rate = math.ceil(empty_batch / total_batch * 100)
            minddata_dict["minddata_device_queue"] = [empty_batch, total_batch, full_batch, total_batch]
            self.__proposal_dict.update(minddata_dict)

        minddata_device_queue_rate = dict()
        for key, value in result.items():
            if isinstance(value, dict):
                status = value.get("status")
                if status == "warning":
                    if key == DATA_PROCESS_WARNING:
                        minddata_device_queue_rate['empty_rate'] = device_queue_empty_rate
                        minddata_device_queue_rate['empty_warning_threshold'] = \
                            self.empty_warning_threshold
                        break
                    elif key == DATA_TRANSMISSION_WARNING:
                        minddata_device_queue_rate['empty_rate'] = device_queue_empty_rate
                        minddata_device_queue_rate['empty_warning_threshold'] = \
                            self.empty_warning_threshold
                        break
                    elif key == DEVICE_QUEUE_WARNING:
                        minddata_dict[key] = ['warning']
                        self.__proposal_dict.update(minddata_dict)
                        break

        if minddata_device_queue_rate:
            minddata_dict["minddata_device_queue_rate"] = minddata_device_queue_rate
            self.__proposal_dict.update(minddata_dict)

    def minddata_cpu_utilization_proposal(self):
        """Get the proposals of minddata cpu utilization"""
        filename = "minddata_cpu_utilization_{}.json".format(self.rank_id)
        file_path = os.path.join(self.profiling_path, filename)
        # Forward compatibility, it is reasonable that the file does not exist.
        if not os.path.exists(file_path):
            return
        minddata_cpu_utilization = OrderedDict()
        minddata_cpu_utilization_analyser = AnalyserFactory.instance().get_analyser(
            'minddata_cpu_utilization', self.profiling_path, self.rank_id)
        try:
            idle_utilization_avg = minddata_cpu_utilization_analyser.get_idle_utilization_avg()
            # The maximum value of this cpu_activate_utilization_avg is 100%.
            cpu_activate_utilization_avg = 100 - idle_utilization_avg
            cpu_activate_utilization_threshold = 80
            if cpu_activate_utilization_avg > cpu_activate_utilization_threshold:
                minddata_cpu_utilization["minddata_cpu_utilization"] = [cpu_activate_utilization_avg]
                self.__proposal_dict.update(minddata_cpu_utilization)
        except (ProfilerRawFileException, ProfilerFileNotFoundException) as err:
            log.exception(err)
